import operator
import typing as ta

from .... import cached
from .... import check
from .... import collections as col
from .... import dataclasses as dc
from .... import lang


KeywordT = ta.TypeVar('KeywordT', bound='Keyword')


##


class Keyword(lang.Abstract):
    tag: ta.ClassVar[str]


##


class KnownKeyword(Keyword, lang.Abstract):
    aliases: ta.ClassVar[frozenset[str]]

    tag_and_aliases: ta.ClassVar[frozenset[str]]

    def __init_subclass__(
            cls,
            *,
            tag: str | None = None,
            aliases: ta.Iterable[str] | None = None,
            **kwargs: ta.Any,
    ) -> None:
        super().__init_subclass__(**kwargs)
        check.empty(set(dir(cls)) & {'tag', 'aliases', 'tag_and_aliases'})
        if not lang.is_abstract_class(cls):
            check.issubclass(cls, lang.Final)
            check.not_isinstance(aliases, str)
            cls.tag = check.non_empty_str(tag)
            cls.aliases = frozenset(aliases or ())
            cls.tag_and_aliases = frozenset([cls.tag, *cls.aliases])
        else:
            for a in (tag, aliases):
                check.none(a)


##


@dc.dataclass(frozen=True)
class Keywords(lang.Final):
    lst: ta.Sequence[Keyword]

    @cached.property
    @dc.init
    def by_type(self) -> ta.Mapping[type[Keyword], Keyword]:
        return col.make_map_by(type, self.lst, strict=True)  # noqa

    @cached.property
    @dc.init
    def by_tag(self) -> ta.Mapping[str, Keyword]:
        return col.make_map_by(operator.attrgetter('tag'), self.lst, strict=True)  # noqa

    def __getitem__(self, item: type[KeywordT] | str) -> KeywordT:
        if isinstance(item, type):
            return self.by_type[item]  # noqa
        elif isinstance(item, str):
            return self.by_tag[item]
        else:
            raise TypeError(item)


##


@dc.dataclass(frozen=True)
class AnyKeyword(Keyword, lang.Abstract):
    v: ta.Any


@dc.dataclass(frozen=True)
class AnyArrayKeyword(Keyword, lang.Abstract):
    vs: ta.Sequence[ta.Any]


@dc.dataclass(frozen=True)
class BooleanKeyword(Keyword, lang.Abstract):
    b: bool


@dc.dataclass(frozen=True)
class NumberKeyword(Keyword, lang.Abstract):
    n: int | float


@dc.dataclass(frozen=True)
class StrKeyword(Keyword, lang.Abstract):
    s: str


@dc.dataclass(frozen=True)
class StrOrStrArrayKeyword(Keyword, lang.Abstract):
    ss: str | ta.Sequence[str]


@dc.dataclass(frozen=True)
class KeywordsKeyword(Keyword, lang.Abstract):
    kw: Keywords


@dc.dataclass(frozen=True)
class KeywordsArrayKeyword(Keyword, lang.Abstract):
    kws: ta.Sequence[Keywords]


@dc.dataclass(frozen=True)
class StrToKeywordsKeyword(Keyword, lang.Abstract):
    m: ta.Mapping[str, Keywords]


@dc.dataclass(frozen=True)
class BooleanOrKeywordsKeyword(Keyword, lang.Abstract):
    bk: bool | Keywords



from .iterators import (  # noqa
    PeekIterator,
    PrefetchIterator,
    ProxyIterator,
    RetainIterator,
)

from .recipes import (  # noqa
    sliding_window,
)

from .tools import (  # noqa
    chunk,
    expand_indexed_pairs,
    merge_on,
    take,
    unzip,
)

from .unique import (  # noqa
    UniqueItem,
    UniqueIterator,
    UniqueStats,
)



import io
import subprocess
import sys
import typing as ta

import anyio.abc

from ... import check
from ...lite.timeouts import Timeout
from ...subprocesses.async_ import AbstractAsyncSubprocesses
from ...subprocesses.run import SubprocessRun
from ...subprocesses.run import SubprocessRunOutput


T = ta.TypeVar('T')


##


class AnyioSubprocesses(AbstractAsyncSubprocesses):
    async def run_(self, run: SubprocessRun) -> SubprocessRunOutput:
        kwargs = dict(run.kwargs or {})

        if run.capture_output:
            kwargs.setdefault('stdout', subprocess.PIPE)
            kwargs.setdefault('stderr', subprocess.PIPE)

        with anyio.fail_after(Timeout.of(run.timeout).or_(None)):
            async with await anyio.open_process(
                    run.cmd,
                    **kwargs,
            ) as proc:
                async def read_output(stream: anyio.abc.ByteReceiveStream, writer: ta.IO) -> None:
                    while True:
                        try:
                            data = await stream.receive()
                        except anyio.EndOfStream:
                            return
                        writer.write(data)

                stdout: io.BytesIO | None = None
                stderr: io.BytesIO | None = None
                async with anyio.create_task_group() as tg:
                    if proc.stdout is not None:
                        stdout = io.BytesIO()
                        tg.start_soon(read_output, proc.stdout, stdout)

                    if proc.stderr is not None:
                        stderr = io.BytesIO()
                        tg.start_soon(read_output, proc.stderr, stderr)

                    if proc.stdin and run.input is not None:
                        await proc.stdin.send(run.input)
                        await proc.stdin.aclose()

                    await proc.wait()

        if run.check and proc.returncode != 0:
            raise subprocess.CalledProcessError(
                ta.cast(int, proc.returncode),
                run.cmd,
                stdout.getvalue() if stdout is not None else None,
                stderr.getvalue() if stderr is not None else None,
            )

        return SubprocessRunOutput(
            proc=proc,

            returncode=check.isinstance(proc.returncode, int),

            stdout=stdout.getvalue() if stdout is not None else None,
            stderr=stderr.getvalue() if stderr is not None else None,
        )

    async def check_call(self, *cmd: str, stdout: ta.Any = sys.stderr, **kwargs: ta.Any) -> None:
        with self.prepare_and_wrap(*cmd, stdout=stdout, check=True, **kwargs) as (cmd, kwargs):  # noqa
            await self.run(*cmd, **kwargs)

    async def check_output(self, *cmd: str, **kwargs: ta.Any) -> bytes:
        with self.prepare_and_wrap(*cmd, stdout=subprocess.PIPE, check=True, **kwargs) as (cmd, kwargs):  # noqa
            return check.not_none((await self.run(*cmd, **kwargs)).stdout)


anyio_subprocesses = AnyioSubprocesses()



import traceback
import typing as ta
import warnings

from ... import lang


##


DEBUG = __debug__


def get_resource_debug() -> bool:
    return DEBUG


def set_resource_debug(debug: bool) -> None:
    global DEBUG
    DEBUG = debug


##


class UnclosedResourceWarning(Warning):
    pass


class Closer(lang.Abstract):
    def __init__(self, *args: ta.Any, **kwargs: ta.Any) -> None:
        super().__init__(*args, **kwargs)

        self.__closed = False

        self.__debug = DEBUG
        if self.__debug:
            self.__repr = repr(self)
            self.__traceback = traceback.format_stack()[:-1]

    @property
    def _is_resourceless(self) -> bool:
        return False

    @ta.final
    def close(self) -> None:
        self.__closed = True
        self._close()

    def _close(self) -> None:
        pass

    def __del__(self) -> None:
        if (
                not self._is_resourceless and
                self.__debug and
                not self.__closed
        ):
            warnings.warn(
                f'\n\n{(sep := ("=" * 40))}\n'
                f'{self.__class__.__name__} object {self.__repr} '
                f'was not properly closed before deletion. Please ensure that `close()` is called before the object is '
                f'deleted.'
                f'\n\n'
                f'{"".join(self.__traceback).rstrip()}'
                f'\n{sep}\n',
                UnclosedResourceWarning,
            )


##


class ResourceNotEnteredError(Exception):
    pass


class ContextCloser(Closer):
    def __init__(self, *args: ta.Any, **kwargs: ta.Any) -> None:
        super().__init__(*args, **kwargs)

        self.__entered = False

    @ta.final
    def __enter__(self) -> ta.Self:
        self.__entered = True
        self._enter()
        return self

    def _enter(self) -> None:
        pass

    def _is_entered(self) -> bool:
        return self.__entered

    def _check_entered(self) -> None:
        if not self.__entered:
            try:
                raise ResourceNotEnteredError(self)  # noqa
            except Exception:
                self.close()
                raise

    @ta.final
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



import sqlalchemy as sa
from sqlalchemy.dialects import sqlite as sal


class SqleanDialect(sal.dialect):  # type: ignore
    name = 'sqlite__sqlean'
    driver = 'sqlean_engine'

    supports_statement_cache = True

    @classmethod
    def import_dbapi(cls):
        from sqlean import dbapi2
        return dbapi2


sa.dialects.registry.register(SqleanDialect.name, __name__, SqleanDialect.__name__)



"""
TODO:
 - collect init_fn's / validate_fns from superclass ClassSpecs
"""
import inspect
import typing as ta

from .... import check
from .... import lang
from ...internals import STD_FIELDS_ATTR
from ...internals import STD_PARAMS_ATTR
from ...processing.driving import drive_cls_processing
from ...specs import ClassSpec
from ...utils import class_decorator
from ..fields.building import build_cls_std_fields
from ..fields.conversion import std_field_to_field_spec
from .metadata import extract_cls_metadata
from .params import STD_PARAM_DEFAULTS
from .params import build_spec_std_params
from .params import get_class_spec


##


@class_decorator
def dataclass(
        cls=None,
        /,
        *,

        init=True,
        repr=True,  # noqa
        eq=True,
        order=False,
        unsafe_hash=False,
        frozen=False,

        match_args=True,
        kw_only=False,
        slots=False,
        weakref_slot=False,

        #

        metadata: ta.Sequence[ta.Any] | None = None,

        reorder: bool | None = None,
        cache_hash: bool | None = None,
        generic_init: bool | None = None,
        override: bool | None = None,

        repr_id: bool | None = None,
        terse_repr: bool | None = None,

        allow_redundant_decorator: bool | None = None,

        #

        _plan_only: bool = False,

        #

        _kwargs: ta.Mapping[str, ta.Any] | None = None,
):
    if isinstance(metadata, ta.Mapping):
        raise TypeError(metadata)

    #

    cls = check.not_none(cls)

    if STD_PARAMS_ATTR in cls.__dict__:
        if (xcs := get_class_spec(cls)) is not None and xcs.allow_redundant_decorator:
            xpd = {
                a: getattr(xcs, a)
                for a in STD_PARAM_DEFAULTS
            }

            dpd = {
                **STD_PARAM_DEFAULTS,
                **{k: v for k, v in (_kwargs or {}).items() if k in STD_PARAM_DEFAULTS},
            }

            if xpd == dpd:
                check.in_(STD_FIELDS_ATTR, cls.__dict__)
                return cls

        raise TypeError(cls)

    check.not_in(STD_PARAMS_ATTR, cls.__dict__)
    check.not_in(STD_FIELDS_ATTR, cls.__dict__)

    #

    fields = build_cls_std_fields(
        cls,
        kw_only=kw_only,
    )
    setattr(cls, STD_FIELDS_ATTR, fields)

    #

    fsl = [
        std_field_to_field_spec(
            f,
            set_metadata=True,
        )
        for f in fields.values()
    ]

    #

    cmd = extract_cls_metadata(cls, deep=True)

    vfp_lst: list[ClassSpec.ValidateFnWithParams] = []
    for md_vf in cmd.validate_fns or []:
        if isinstance(md_vf, staticmethod):
            md_vf = md_vf.__func__
        vfp_lst.append(ClassSpec.ValidateFnWithParams(
            md_vf,
            [p.name for p in inspect.signature(md_vf).parameters.values()],
        ))

    #

    cs = ClassSpec(
        fields=fsl,

        init=init,
        repr=repr,
        eq=eq,
        order=order,
        unsafe_hash=unsafe_hash,
        frozen=frozen,

        match_args=match_args,
        kw_only=kw_only,
        slots=slots,
        weakref_slot=weakref_slot,

        #

        metadata=(
            *(cmd.user_metadata or []),
            *(metadata or []),
        ) or None,

        **{
            **(cmd.extra_params or {}),
            **lang.opt_kw(
                reorder=reorder,
                cache_hash=cache_hash,
                generic_init=generic_init,
                override=override,

                repr_id=repr_id,
                terse_repr=terse_repr,

                allow_redundant_decorator=allow_redundant_decorator,
            ),
        },

        init_fns=cmd.init_fns or None,
        validate_fns=vfp_lst or None,
    )

    #

    std_params = build_spec_std_params(cs)
    setattr(cls, STD_PARAMS_ATTR, std_params)

    #

    return drive_cls_processing(
        cls,
        cs,
        plan_only=_plan_only,
    )



from .base import (  # noqa
    Kv,
    MutableKv,
)

from .capabilities import (  # noqa
    Closeable,
    close,
    closing,

    Flushable,
    flush,
)

from .filtered import (  # noqa
    KeyFilteredKv,
    KeyFilteredMutableKv,

    ValueFilteredKeyError,
    ValueFilteredKv,
    ValueFilteredMutableKv,
)

from .mappings import (  # noqa
    MappingKv,
    MappingMutableKv,

    KvMapping,
    KvMutableMapping,
)

from .transformed import (  # noqa
    KeyTransformedKv,
    KeyTransformedMutableKey,

    ValueTransformedKv,
    ValueTransformedMutableKv,
)

from .wrappers import (  # noqa
    WrapperKv,

    underlying,
    underlying_of,

    SimpleWrapperKv,
    SimpleWrapperMutableKv,

    UnmodifiableError,
    UnmodifiableKv,
)



# ruff: noqa: Q000
"""
https:#github.com/golang/go/blob/3d33437c450aa74014ea1d41cd986b6ee6266984/src/strconv/quote.go
"""
# Copyright 2009 The Go Authors.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#      disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#      following disclaimer in the documentation and/or other materials provided with the distribution.
#    * Neither the name of Google LLC nor the names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


def unhex(b: str) -> int | None:
    c = ord(b)
    if ord('0') <= c <= ord('9'):
        return c - ord('0')
    elif ord('a') <= c <= ord('f'):
        return c - ord('a') + 10
    elif ord('A') <= c <= ord('F'):
        return c - ord('A') + 10
    else:
        return None


SURROGATE_MIN = 0xD800
SURROGATE_MAX = 0xDFFF
MAX_RUNE = ord('\U0010FFFF')  # Maximum valid Unicode code point.


def is_valid_utf8(r: int) -> bool:
    if 0 <= r < SURROGATE_MIN:
        return True
    elif SURROGATE_MAX < r <= MAX_RUNE:
        return True
    return False


##


class UnquoteError(Exception):
    pass


def unquote_char(s: str, quote: str) -> tuple[str, bool, str]:  # (value, multibyte, tail)
    # UnquoteChar decodes the first character or byte in the escaped string or character literal represented by the
    # string s. It returns four values:
    #
    #  1. value, the decoded Unicode code point or byte value;
    #  2. multibyte, a boolean indicating whether the decoded character requires a multibyte UTF-8 representation;
    #  3. tail, the remainder of the string after the character; and
    #  4. an error that will be nil if the character is syntactically valid.
    #
    # The second argument, quote, specifies the type of literal being parsed and therefore which escaped quote character
    # is permitted.
    # If set to a single quote, it permits the sequence \' and disallows unescaped '.
    # If set to a double quote, it permits \" and disallows unescaped ".
    # If set to zero, it does not permit either escape and allows both quote characters to appear unescaped.

    # easy cases
    if not s:
        raise UnquoteError

    c = s[0]
    if c == quote and quote in '\'"':
        raise UnquoteError
    # elif c >= utf8.graphic_only:
    #     r, size = utf8.DecodeRuneInString(s)
    #     return r, True, s[size:]
    elif c != '\\':
        return s[0], False, s[1:]

    # hard case: c is backslash
    if len(s) <= 1:
        raise UnquoteError

    c = s[1]
    s = s[2:]
    value: str
    multibyte = False

    if c == 'a':
        value = '\a'
    elif c == 'b':
        value = '\b'
    elif c == 'f':
        value = '\f'
    elif c == 'n':
        value = '\n'
    elif c == 'r':
        value = '\r'
    elif c == 't':
        value = '\t'
    elif c == 'v':
        value = '\v'
    elif c in 'xuU':
        n = 0
        if c == 'x':
            n = 2
        elif c == 'u':
            n = 4
        elif c == 'U':
            n = 8
        if len(s) < n:
            raise UnquoteError
        v = 0
        for j in range(n):
            x = unhex(s[j])
            if x is None:
                raise UnquoteError
            v = v << 4 | x
        s = s[n:]
        if c == 'x':
            try:
                value = chr(v)  # noqa
            except ValueError:
                raise UnquoteError from None
        else:
            # single-byte string, possibly not UTF-8
            if not is_valid_utf8(v):
                raise UnquoteError
            try:
                value = chr(v)  # noqa
            except ValueError:
                raise UnquoteError from None
            multibyte = True
    elif c in '01234567':
        v = ord(c) - ord('0')
        if len(s) < 2:
            raise UnquoteError
        for j in range(2):  # one digit already; two more
            x = ord(s[j]) - ord('0')
            if x < 0 or x > 7:
                raise UnquoteError
            v = (v << 3) | x
        s = s[2:]
        if v > 255:
            raise UnquoteError
        value = chr(v)
    elif c == '\\':
        value = '\\'
    elif c in ('\'', '"'):
        if c != quote:
            raise UnquoteError
        value = c
    else:
        raise UnquoteError

    tail = s
    return value, multibyte, tail


def unquote_(ins: str, unescape: bool) -> tuple[str, str]:  # (out, rem)
    # unquote parses a quoted string at the start of the input, returning the parsed prefix, the remaining suffix, and
    # any parse errors. If unescape is true, the parsed prefix is unescaped, otherwise the input prefix is provided
    # verbatim.

    # Determine the quote form and optimistically find the terminating quote.
    if len(ins) < 2:
        raise UnquoteError
    quote = ins[0]
    end = ins[1:].find(quote)
    if end < 0:
        raise UnquoteError
    end += 2  # position after terminating quote; may be wrong if escape sequences are present

    if quote == '`':
        if not unescape:
            out = ins[:end]  # include quotes
        elif '\r' not in ins[:end]:
            out = ins[len("`"): end - len("`")]  # exclude quotes
        else:
            # Carriage return characters ('\r') inside raw string literals are discarded from the raw string value.
            buf = []
            for i in range(len("`"), end - len("`")):
                if ins[i] != '\r':
                    buf.append(ins[i])
            out = ''.join(buf)

        # NOTE: Prior implementations did not verify that raw strings consist of valid UTF-8 characters and we continue
        # to not verify it as such. The Go specification does not explicitly require valid UTF-8, but only mention that
        # it is implicitly valid for Go source code (which must be valid UTF-8).
        return out, ins[end:]

    elif quote in ('"', '\''):
        # Handle quoted strings without any escape sequences.
        if '\\' not in ins[:end] and '\n' not in ins[:end]:
            valid = False
            if quote == '"':
                # valid = utf8.ValidString(ins[len('"'): end - len('"')])
                valid = True
            elif quote == '\'':
                # r, n = utf8.DecodeRuneInString(ins[len("'"): end - len("'")])
                # valid = len("'") + n + len("'") == end and (r != utf8.RuneError or n != 1)
                valid = end == 3
            if valid:
                out = ins[:end]
                if unescape:
                    out = out[1: end - 1]  # exclude quotes
                return out, ins[end:]

        # Handle quoted strings with escape sequences.
        buf = []
        in0 = ins
        ins = ins[1:]  # skip starting quote

        while ins and ins[0] != quote:
            # Process the next character, rejecting any unescaped newline characters which are invalid.
            r, multibyte, rem = unquote_char(ins, quote)
            if ins[0] == '\n':
                raise UnquoteError
            ins = rem

            # Append the character if unescaping the input.
            if unescape:
                # if r < utf8.RuneSelf or not multibyte:
                #     buf.append(byte(r))
                # else:
                buf.append(r)

            # Single quoted strings must be a single character.
            if quote == '\'':
                break

        # Verify that the string ends with a terminating quote.
        if not (ins and ins[0] == quote):
            raise UnquoteError

        ins = ins[1:]  # skip terminating quote

        if unescape:
            return ''.join(buf), ins

        return in0[:len(in0) - len(ins)], ins

    else:
        raise UnquoteError


def unquote(s: str) -> str:
    # Unquote interprets s as a single-quoted, double-quoted, or backquoted Go string literal, returning the string
    # value that s quotes.  (If s is single-quoted, it would be a Go character literal; Unquote returns the
    # corresponding one-character string.)
    out, rem = unquote_(s, True)
    if rem:
        raise UnquoteError
    return out



# ruff: noqa: UP006 UP007
# @omlish-lite
import dataclasses as dc
import re
import typing as ta

from ..lite.cached import cached_nullary
from ..lite.check import check


##


@dc.dataclass(frozen=True)
class StringMangler:
    escape: str
    escaped: ta.Sequence[str]

    @classmethod
    def of(cls, escape: str, escaped: ta.Iterable[str]) -> 'StringMangler':
        check.arg(len(escape) == 1)
        return StringMangler(escape, sorted(set(escaped) - {escape}))

    def __post_init__(self) -> None:
        check.non_empty_str(self.escape)
        check.arg(len(self.escape) == 1)
        check.not_in(self.escape, self.escaped)
        check.arg(len(set(self.escaped)) == len(self.escaped))

    #

    @cached_nullary
    def replaced(self) -> ta.Tuple[str, ...]:
        return (self.escape, *self.escaped)

    @cached_nullary
    def replaced_set(self) -> ta.FrozenSet[str]:
        return frozenset(self.replaced())

    @cached_nullary
    def replaced_indexes(self) -> ta.Mapping[str, int]:
        return {s: i for i, s in enumerate(self.replaced())}

    @cached_nullary
    def replaced_pat(self) -> re.Pattern:
        return re.compile('|'.join(re.escape(k) for k in self.replaced()))

    #

    @cached_nullary
    def replacement_pad(self) -> int:
        return len('%x' % (len(self.replaced()),))  # noqa

    @cached_nullary
    def replacements(self) -> ta.Sequence[ta.Tuple[str, str]]:
        fmt = f'%0{self.replacement_pad()}x'
        return [
            (l, self.escape + fmt % (i,))
            for i, l in enumerate(self.replaced())
        ]

    @cached_nullary
    def replacements_dict(self) -> ta.Mapping[str, str]:
        return dict(self.replacements())

    @cached_nullary
    def inverse_replacements_dict(self) -> ta.Mapping[str, str]:
        return {v: k for k, v in self.replacements()}

    @cached_nullary
    def replacements_pat(self) -> re.Pattern:
        return re.compile(''.join([re.escape(self.escape), '.' * self.replacement_pad()]))

    #

    # def mangle(self, s: str) -> str:
    #     ecs = sorted(
    #         frozenset(s) & self.replaced_set(),
    #         key=self.replaced_indexes().__getitem__,
    #         )
    #     rd = self.replacements_dict()
    #     for l in ecs:
    #         r = rd[l]
    #         s = s.replace(l, r)
    #     return s

    def mangle(self, s: str) -> str:
        rd = self.replacements_dict()
        return self.replaced_pat().sub(lambda m: rd[m.group(0)], s)

    #

    # def unmangle(self, s: str) -> str:
    #     for l, r in reversed(self.replacements()):
    #         s = s.replace(r, l)
    #     return s

    def unmangle(self, s: str) -> str:
        ird = self.inverse_replacements_dict()
        return self.replacements_pat().sub(lambda m: ird[m.group(0)], s)



import html
import io
import subprocess
import tempfile
import time
import typing as ta

from ... import dispatch
from .items import Attrs
from .items import Cell
from .items import Edge
from .items import Graph
from .items import Id
from .items import Item
from .items import Node
from .items import Raw
from .items import RawStmt
from .items import Row
from .items import Table
from .items import Text


class Renderer:
    def __init__(self, out: ta.TextIO) -> None:
        super().__init__()

        self._out = out

    @dispatch.method
    def render(self, item: Item) -> None:
        raise TypeError(item)

    @render.register
    def render_raw(self, item: Raw) -> None:
        self._out.write(item.raw)

    @render.register
    def render_text(self, item: Text) -> None:
        self._out.write(html.escape(item.text))

    @render.register
    def render_cell(self, item: Cell) -> None:
        self._out.write('<td>')
        self.render(item.value)
        self._out.write('</td>')

    @render.register
    def render_row(self, item: Row) -> None:
        self._out.write('<tr>')
        for cell in item.cells:
            self.render(cell)
        self._out.write('</tr>')

    @render.register
    def render_table(self, item: Table) -> None:
        self._out.write('<table>')
        for row in item.rows:
            self.render(row)
        self._out.write('</table>')

    @render.register
    def render_id(self, item: Id) -> None:
        self._out.write(f'"{item.id}"')

    @render.register
    def render_attrs(self, item: Attrs) -> None:
        if item.attrs:
            self._out.write('[')
            for i, (k, v) in enumerate(item.attrs.items()):
                if i:
                    self._out.write(', ')
                self._out.write(k)
                self._out.write('=<')
                self.render(v)
                self._out.write('>')
            self._out.write(']')

    @render.register
    def render_raw_stmt(self, item: RawStmt) -> None:
        self._out.write(item.raw)
        self._out.write('\n')

    @render.register
    def render_edge(self, item: Edge) -> None:
        self.render(item.left)
        self._out.write(' -> ')
        self.render(item.right)
        if item.attrs.attrs:
            self._out.write(' ')
            self.render(item.attrs)
        self._out.write(';\n')

    @render.register
    def render_node(self, item: Node) -> None:
        self.render(item.id)
        if item.attrs.attrs:
            self._out.write(' ')
            self.render(item.attrs)
        self._out.write(';\n')

    @render.register
    def render_graph(self, item: Graph) -> None:
        self._out.write('digraph ')
        self.render(item.id)
        self._out.write(' {\n')
        for stmt in item.stmts:
            self.render(stmt)
        self._out.write('}\n')


def render(item: Item) -> str:
    out = io.StringIO()
    Renderer(out).render(item)
    return out.getvalue()


def open_dot(
        gv: str,
        *,
        timeout_s: float = 1.,
        sleep_s: float = 0.,
        delete: bool = False,
) -> None:
    stdout, _ = subprocess.Popen(
        ['dot', '-Tpdf'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
    ).communicate(
        input=gv.encode('utf-8'),
        timeout=timeout_s,
    )

    with tempfile.NamedTemporaryFile(
        suffix='.pdf',
        delete=delete,
    ) as pdf:
        pdf.file.write(stdout)
        pdf.file.flush()

    _, _ = subprocess.Popen(
        ['open', pdf.name],
    ).communicate(
        timeout=timeout_s,
    )

    if sleep_s > 0.:
        time.sleep(sleep_s)



"""
TODO:
 - generics? injected Manifests / type contexts?
 - ta.Literal? support for simple scalars, enums
 - multidispatch? never solved..
  - just generic on tuple[A0, A1, ...]
"""
import inspect
import typing as ta

from .. import c3
from .. import check
from .. import reflect as rfl


T = ta.TypeVar('T')


##


def get_impl_func_cls_set(func: ta.Callable, *, arg_offset: int = 0) -> frozenset[type]:
    ann = getattr(func, '__annotations__', {})
    if not ann:
        raise TypeError(f'Invalid impl func: {func!r}')

    def erase(a):
        if isinstance(a, rfl.Generic):
            return a.cls
        else:
            return check.isinstance(a, type)

    # Exclude 'return' to support difficult to handle return types - they are unimportant.
    th_dct = rfl.get_filtered_type_hints(func, exclude=['return'])

    ps = inspect.signature(func).parameters
    p = list(ps.values())[arg_offset]
    cls = th_dct[p.name]

    rty = rfl.type_(cls)
    if isinstance(rty, rfl.Union):
        ret = frozenset(erase(arg) for arg in rty.args)
    else:
        ret = frozenset([erase(rty)])

    return ret


##


def find_impl(cls: type, registry: ta.Mapping[type, T]) -> T | None:
    mro = c3.compose_mro(cls, registry.keys())

    match: type | None = None
    for t in mro:
        if match is not None:
            # If *match* is an implicit ABC but there is another unrelated, equally matching implicit ABC, refuse the
            # temptation to guess.
            if (
                    t in registry
                    and t not in cls.__mro__
                    and match not in cls.__mro__
                    and not issubclass(match, t)
            ):
                raise RuntimeError(f'Ambiguous dispatch: {match} or {t}')
            break

        if t in registry:
            match = t

    if match is None:
        return None
    return registry.get(match)



import abc
import typing as ta

from ... import lang


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class PersistentMap(lang.Abstract, ta.Generic[K, V]):
    @abc.abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def __contains__(self, item: K) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def __getitem__(self, item: K) -> V:
        raise NotImplementedError

    @abc.abstractmethod
    def __iter__(self) -> ta.Iterator[K]:
        raise NotImplementedError

    @abc.abstractmethod
    def items(self) -> ta.Iterator[tuple[K, V]]:
        raise NotImplementedError

    @abc.abstractmethod
    def with_(self, k: K, v: V) -> ta.Self:
        raise NotImplementedError

    @abc.abstractmethod
    def without(self, k: K) -> ta.Self:
        raise NotImplementedError

    @abc.abstractmethod
    def default(self, k: K, v: V) -> ta.Self:
        raise NotImplementedError


class PersistentMapping(
    PersistentMap[K, V],
    ta.Mapping[K, V],
    lang.Abstract,
    ta.Generic[K, V],
):
    @abc.abstractmethod
    def __contains__(self, item: K) -> bool:  # type: ignore[override]
        raise NotImplementedError

    @abc.abstractmethod
    def items(self) -> ta.Iterator[tuple[K, V]]:  # type: ignore[override]  # FIXME: ItemsView
        raise NotImplementedError



# ruff: noqa: I001
import typing as _ta

from ... import lang as _lang


if _ta.TYPE_CHECKING:
    from .apiadapter import (  # noqa
        SqlalchemyApiWrapper,

        SqlalchemyApiRows,
        SqlalchemyApiConn,
        SqlalchemyApiDb,
        SqlalchemyApiAdapter,

        api_adapt,
    )

else:
    _lang.proxy_init(globals(), '.apiadapter', [
        'SqlalchemyApiWrapper',

        'SqlalchemyApiRows',
        'SqlalchemyApiConn',
        'SqlalchemyApiDb',
        'SqlalchemyApiAdapter',

        'api_adapt',
    ])


from .asyncs import (  # noqa
    AsyncConnection,
    AsyncConnectionLike,
    AsyncEngine,
    AsyncEngineLike,
    AsyncTransaction,
    AsyncTransactionLike,
    async_adapt,
)

from .exprs import (  # noqa
    paren,
)



import dataclasses as dc
import typing as ta
import weakref

from .. import cached
from .. import check
from .. import collections as col
from .. import lang
from .. import reflect as rfl
from .internals import STD_FIELDS_ATTR
from .internals import StdFieldType
from .internals import std_field_type
from .internals import std_is_kw_only


ClassAnnotations: ta.TypeAlias = ta.Mapping[str, ta.Any]


##


def _get_cls_annotations(cls: type) -> ClassAnnotations:
    # Does not use ta.get_type_hints because that's what std dataclasses do [1]. Might be worth revisiting? A part
    # of why they don't is to not import typing for efficiency but we don't care about that degree of startup speed.
    # [1]: https://github.com/python/cpython/blob/54c63a32d06cb5f07a66245c375eac7d7efb964a/Lib/dataclasses.py#L985-L986  # noqa
    return rfl.get_annotations(cls)


_CLS_ANNOTATIONS_CACHE: ta.MutableMapping[type, ClassAnnotations] = weakref.WeakKeyDictionary()


def get_cls_annotations(cls: type) -> ClassAnnotations:
    try:
        return _CLS_ANNOTATIONS_CACHE[cls]
    except KeyError:
        pass
    ret = _get_cls_annotations(cls)
    _CLS_ANNOTATIONS_CACHE[cls] = ret
    return ret


##


class FieldsInspection:
    def __init__(
            self,
            cls: type,
            *,
            cls_fields: ta.Mapping[str, dc.Field] | None = None,
    ) -> None:
        super().__init__()

        self._cls = cls

        self._given_cls_fields = cls_fields
        if cls_fields is None:
            cls_fields = getattr(self._cls, STD_FIELDS_ATTR)
        self._cls_fields = cls_fields

    @property
    def cls(self) -> type:
        return self._cls  # noqa

    @property
    def cls_fields(self) -> ta.Mapping[str, dc.Field]:
        return self._cls_fields

    @cached.property
    def cls_annotations(self) -> ta.Mapping[str, ta.Any]:
        return get_cls_annotations(self._cls)

    #

    class _FoundFields(ta.NamedTuple):
        fields: dict[str, dc.Field]
        field_owners: dict[str, type]

    @lang.cached_function(no_wrapper_update=True)
    def _find_fields(self) -> _FoundFields:
        fields: dict[str, dc.Field] = {}
        field_owners: dict[str, type] = {}

        for b in self._cls.__mro__[-1:0:-1]:
            base_fields = getattr(b, STD_FIELDS_ATTR, None)
            if base_fields is not None:
                for name in get_cls_annotations(b):
                    try:
                        f = base_fields[name]
                    except KeyError:
                        continue
                    fields[f.name] = f
                    field_owners[f.name] = b

        cls_fields = self._cls_fields
        for name, ann in self.cls_annotations.items():
            if std_is_kw_only(self._cls, ann):
                continue
            fields[name] = cls_fields[name]
            field_owners[name] = self._cls

        return FieldsInspection._FoundFields(fields, field_owners)

    @cached.property
    def fields(self) -> ta.Mapping[str, dc.Field]:
        return self._find_fields().fields

    @cached.property
    def instance_fields(self) -> ta.Sequence[dc.Field]:
        return [f for f in self.fields.values() if std_field_type(f) is StdFieldType.INSTANCE]

    @cached.property
    def field_owners(self) -> ta.Mapping[str, type]:
        return self._find_fields().field_owners

    ##

    @cached.property
    def generic_mro(self) -> ta.Sequence[rfl.Type]:
        return rfl.ALIAS_UPDATING_GENERIC_SUBSTITUTION.generic_mro(self._cls)

    @cached.property
    def generic_mro_lookup(self) -> ta.Mapping[type, rfl.Type]:
        return col.make_map(((check.not_none(rfl.get_concrete_type(g)), g) for g in self.generic_mro), strict=True)

    def generic_replaced_field_type(self, fn: str) -> rfl.Type:
        f = self.fields[fn]
        fo = self.field_owners[f.name]
        go = self.generic_mro_lookup[fo]
        tvr = rfl.get_type_var_replacements(go)
        fty = rfl.type_(f.type)
        rty = rfl.replace_type_vars(fty, tvr, update_aliases=True)
        return rty

    @cached.property
    def generic_replaced_field_types(self) -> ta.Mapping[str, rfl.Type]:
        ret: dict[str, ta.Any] = {}
        for f in self.fields.values():
            ret[f.name] = self.generic_replaced_field_type(f.name)
        return ret

    @cached.property
    def generic_replaced_field_annotations(self) -> ta.Mapping[str, ta.Any]:
        return {k: rfl.to_annotation(v) for k, v in self.generic_replaced_field_types.items()}


def inspect_fields(cls: type) -> FieldsInspection:
    return FieldsInspection(cls)



"""
TODO:
 - inline polymorph: 'ami: foo' -> 'image: {'ami': {'name': 'foo'}}
 - parse duration and size settings, "512k" or "10 seconds"
 - includes
 - substitutions ("foo" : ${bar}, "foo" : Hello ${who})
"""



"""
TODO:
 - 'setup' locals - for FrozenGenerator's `condition += ' or name in {' + ', '.join(repr(f.name) for f in fields) + '}'`
   allow for `condition += ' or name in {fields_frozenset_ident}' with fields_frozenset_ident setup in preamble
"""
import sys
import types
import typing as ta

from ... import lang
from ..utils import repr_round_trip_value
from .globals import FN_GLOBAL_VALUES
from .globals import FnGlobal
from .idents import CLS_IDENT
from .ops import AddMethodOp
from .ops import AddPropertyOp
from .ops import IfAttrPresent
from .ops import Op
from .ops import OpRef
from .ops import OpRefMap
from .ops import Ref
from .ops import SetAttrOp


T = ta.TypeVar('T')


##


class OpExecutor:
    def __init__(self, cls: type, orm: OpRefMap) -> None:
        super().__init__()

        self._cls = cls
        self._orm = orm

    #

    @lang.cached_function(no_wrapper_update=True)
    def _cls_globals(self) -> dict[str, ta.Any]:
        if self._cls.__module__ in sys.modules:
            return sys.modules[self._cls.__module__].__dict__
        else:
            return {}

    #

    def _set_fn_qualname(self, fn: types.FunctionType, name: str | None = None) -> types.FunctionType:
        if name is None:
            name = fn.__name__
        fn.__qualname__ = f'{self._cls.__qualname__}.{name}'
        return fn

    def _create_fn(
            self,
            name: str,
            src: str,
            refs: ta.Iterable[Ref] = (),
    ) -> types.FunctionType:
        lo: dict = {
            CLS_IDENT: self._cls,
            **FN_GLOBAL_VALUES,
        }
        for r in refs:
            if isinstance(r, OpRef):
                lo[r.ident()] = self._orm[r]
            elif isinstance(r, FnGlobal):
                pass
            else:
                raise TypeError(r)

        x_src = '\n'.join([
            f'def __create_fn__(',
            f'    *,',
            *[
                f'    {k},'
                for k in lo
            ],
            f'):',
            *[
                f'    {l}'
                for l in src.splitlines()
            ],
            f'',
            f'    return {name}',
        ])

        ns: dict = {}
        exec(
            x_src,
            self._cls_globals(),
            ns,
        )
        x_fn = ns['__create_fn__']
        fn = x_fn(**lo)

        if not isinstance(fn, types.FunctionType):
            raise TypeError(fn)

        self._set_fn_qualname(fn, name)

        return fn

    def _create_opt_fn(
            self,
            name: str,
            src: str | None,
            *args: ta.Any,
            **kwargs: ta.Any,
    ) -> types.FunctionType | None:
        if src is None:
            return None

        return self._create_fn(
            name,
            src,
            *args,
            **kwargs,
        )

    #

    def _execute_set_attr(
            self,
            attr_name: str,
            value: ta.Any,
            if_present: IfAttrPresent,
    ) -> None:
        if attr_name in self._cls.__dict__:
            if if_present == 'skip':
                return
            elif if_present == 'replace':
                pass
            elif if_present == 'error':
                raise TypeError(f'Cannot overwrite attribute {attr_name} in class {self._cls.__name__}')
            else:
                raise ValueError(if_present)

        setattr(self._cls, attr_name, value)

    def execute(self, op: Op) -> None:
        if isinstance(op, SetAttrOp):
            if isinstance(v := op.value, OpRef):
                v = self._orm[v]
                if isinstance(v, types.FunctionType):
                    self._set_fn_qualname(v)
            else:
                v = repr_round_trip_value(v)

            self._execute_set_attr(
                op.name,
                v,
                op.if_present,
            )

        elif isinstance(op, AddMethodOp):
            fn = self._create_fn(
                op.name,
                op.src,
                op.refs,
            )

            self._execute_set_attr(
                op.name,
                fn,
                op.if_present,
            )

        elif isinstance(op, AddPropertyOp):
            get_fn = self._create_opt_fn(
                op.name,
                op.get_src,
                op.refs,
            )
            set_fn = self._create_opt_fn(
                op.name,
                op.set_src,
                op.refs,
            )
            del_fn = self._create_opt_fn(
                op.name,
                op.del_src,
                op.refs,
            )
            prop = property(
                get_fn,
                set_fn,
                del_fn,
            )

            setattr(self._cls, op.name, prop)

        else:
            raise TypeError(op)



import functools
import heapq
import itertools
import typing as ta

from .iterators import PeekIterator
from .iterators import PrefetchIterator


T = ta.TypeVar('T')
U = ta.TypeVar('U')


##


def unzip(it: ta.Iterable[T], width: int | None = None) -> list:
    if width is None:
        if not isinstance(it, PeekIterator):
            it = PeekIterator(iter(it))
        try:
            width = len(it.peek())
        except StopIteration:
            return []

    its: list[PrefetchIterator[T]] = []
    running = True

    def next_fn(idx):
        nonlocal running
        if not running:
            raise StopIteration
        try:
            items = next(it)  # type: ignore
        except StopIteration:
            running = False
            raise
        for item_idx, item in enumerate(items):
            its[item_idx].push(item)
        return next(its[idx])

    its.extend(PrefetchIterator(functools.partial(next_fn, idx)) for idx in range(width))
    return its


def take(n: int, iterable: ta.Iterable[T]) -> list[T]:
    return list(itertools.islice(iterable, n))


def chunk(n: int, iterable: ta.Iterable[T], strict: bool = False) -> ta.Iterator[list[T]]:
    # TODO: remove with 3.13 - 3.12 doesn't support strict
    iterator = iter(functools.partial(take, n, iter(iterable)), [])
    if strict:
        def ret():
            for chunk in iterator:
                if len(chunk) != n:
                    raise ValueError('iterable is not divisible by n.')
                yield chunk
        return iter(ret())
    else:
        return iterator


def merge_on(
        function: ta.Callable[[T], U],
        *its: ta.Iterable[T],
) -> ta.Iterator[tuple[U, list[tuple[int, T]]]]:
    indexed_its = [
        (
            (function(item), it_idx, item)
            for it_idx, item in zip(itertools.repeat(it_idx), it)
        )
        for it_idx, it in enumerate(its)
    ]

    grouped_indexed_its = itertools.groupby(
        heapq.merge(*indexed_its),
        key=lambda item_tuple: item_tuple[0],
    )

    return (
        (fn_item, [(it_idx, item) for _, it_idx, item in grp])
        for fn_item, grp in grouped_indexed_its
    )


def expand_indexed_pairs(
        seq: ta.Iterable[tuple[int, T]],
        default: T,
        *,
        width: int | None = None,
) -> list[T]:
    width_ = width
    if width_ is None:
        width_ = (max(idx for idx, _ in seq) + 1) if seq else 0
    result = [default] * width_
    for idx, value in seq:
        if idx < width_:
            result[idx] = value
    return result



import typing as ta

from ... import check
from ... import lang
from .asquery import as_query
from .base import Querier
from .base import Rows
from .queries import QueryMode
from .rows import Row


##


def exec(  # noqa
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> None:
    q = as_query(
        obj,
        *args,
        mode=QueryMode.EXEC,
        querier=querier,
    )

    with querier.query(q):
        pass


##


def query(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Rows:
    q = as_query(
        obj,
        *args,
        mode=QueryMode.QUERY,
        querier=querier,
    )

    return querier.query(q)


#


def query_all(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> list[Row]:
    with query(querier, obj, *args) as rows:
        return list(rows)


def query_first(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row:
    with query(querier, obj, *args) as rows:
        return next(rows)


def query_opt_first(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row | None:
    with query(querier, obj, *args) as rows:
        return next(rows, None)


def query_one(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row:
    with query(querier, obj, *args) as rows:
        return check.single(rows)


def query_opt_one(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row | None:
    with query(querier, obj, *args) as rows:
        return check.opt_single(rows)


def query_scalar(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Any:
    row = query_one(querier, obj, *args)
    return check.single(row.values)


def query_maybe_scalar(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> lang.Maybe[ta.Any]:
    row = query_opt_one(querier, obj, *args)
    if row is not None:
        return lang.just(check.single(row.values))
    else:
        return lang.empty()



# ruff: noqa: UP006 UP007
# @omlish-lite
import io
import socket
import typing as ta


##


class SocketWriter(io.BufferedIOBase):
    """
    Simple writable BufferedIOBase implementation for a socket

    Does not hold data in a buffer, avoiding any need to call flush().
    """

    def __init__(self, sock):
        super().__init__()

        self._sock = sock

    def writable(self):
        return True

    def write(self, b):
        self._sock.sendall(b)
        with memoryview(b) as view:
            return view.nbytes

    def fileno(self):
        return self._sock.fileno()


class SocketIoPair(ta.NamedTuple):
    r: ta.BinaryIO
    w: ta.BinaryIO

    @classmethod
    def from_socket(
            cls,
            sock: socket.socket,
            *,
            r_buf_size: int = -1,
            w_buf_size: int = 0,
    ) -> 'SocketIoPair':
        rf: ta.Any = sock.makefile('rb', r_buf_size)

        if w_buf_size:
            wf: ta.Any = SocketWriter(sock)
        else:
            wf = sock.makefile('wb', w_buf_size)

        return cls(rf, wf)


##


def close_socket_immediately(sock: socket.socket) -> None:
    try:
        # Explicitly shutdown. socket.close() merely releases the socket and waits for GC to perform the actual close.
        sock.shutdown(socket.SHUT_WR)

    except OSError:
        # Some platforms may raise ENOTCONN here
        pass

    sock.close()



# ruff: noqa: UP006 UP007
import dataclasses as dc
import typing as ta


##


def is_immediate_dataclass(cls: type) -> bool:
    if not isinstance(cls, type):
        raise TypeError(cls)
    return dc._FIELDS in cls.__dict__  # type: ignore[attr-defined]  # noqa


##


def dataclass_cache_hash(
        *,
        cached_hash_attr: str = '__dataclass_hash__',
):
    def inner(cls):
        if not isinstance(cls, type) and dc.is_dataclass(cls):
            raise TypeError(cls)

        if (
                cls.__hash__ is object.__hash__ or
                '__hash__' not in cls.__dict__
        ):
            raise TypeError(cls)

        real_hash = cls.__hash__

        def cached_hash(self) -> int:
            try:
                return object.__getattribute__(self, cached_hash_attr)
            except AttributeError:
                object.__setattr__(self, cached_hash_attr, h := real_hash(self))  # type: ignore[call-arg]
            return h

        cls.__hash__ = cached_hash  # type: ignore[method-assign]

        return cls

    return inner


##


def dataclass_maybe_post_init(sup: ta.Any) -> bool:
    if not isinstance(sup, super):
        raise TypeError(sup)
    try:
        fn = sup.__post_init__  # type: ignore
    except AttributeError:
        return False
    fn()
    return True


##


def dataclass_repr_filtered(
        obj: ta.Any,
        fn: ta.Callable[[ta.Any, dc.Field, ta.Any], bool],
) -> str:
    return (
        f'{obj.__class__.__qualname__}(' +
        ', '.join([
            f'{f.name}={v!r}'
            for f in dc.fields(obj)
            if fn(obj, f, v := getattr(obj, f.name))
        ]) +
        ')'
    )


def dataclass_repr_omit_none(obj: ta.Any) -> str:
    return dataclass_repr_filtered(obj, lambda o, f, v: v is not None)


def dataclass_repr_omit_falsey(obj: ta.Any) -> str:
    return dataclass_repr_filtered(obj, lambda o, f, v: bool(v))



# ruff: noqa: I001
import typing as _ta

from .. import lang as _lang


from .coerce import (  # noqa
    abs_set,
    abs_set_of,
    abs_set_of_or_none,
    abs_set_or_none,
    frozenset_,
    frozenset_of,
    frozenset_of_or_none,
    frozenset_or_none,
    map,  # noqa
    map_of,
    map_of_or_none,
    map_or_none,
    opt_abs_set,
    opt_abs_set_of,
    opt_frozenset,
    opt_frozenset_of,
    opt_map,
    opt_map_of,
    opt_seq,
    opt_seq_of,
    seq,
    seq_of,
    seq_of_or_none,
    seq_or_none,
)

from .errors import (  # noqa
    DuplicateKeyError,
)

from .frozen import (  # noqa
    Frozen,
    FrozenDict,
    FrozenList,
    frozendict,
    frozenlist,
)

from .hasheq import (  # noqa
    HashEq,
    HashEqMap,
    HashEq_,
    hash_eq,
)

from .identity import (  # noqa
    IdentityKeyDict,
    IdentitySet,
    IdentityWrapper,
    IdentityWeakSet,
)

if _ta.TYPE_CHECKING:
    from . import kv
else:
    kv = _lang.proxy_import('.kv', __package__)

from .mappings import (  # noqa
    MissingDict,
    TypeMap,
    DynamicTypeMap,
    guarded_map_update,
    multikey_dict,
)

from .ordered import (  # noqa
    OrderedFrozenSet,
    OrderedSet,
)

from .persistent.persistent import (  # noqa
    PersistentMap,
    PersistentMapping,
)

if _ta.TYPE_CHECKING:
    from .persistent.treapmap import (  # noqa
        TreapDict,
        TreapMap,
        new_treap_dict,
        new_treap_map,
    )
else:
    _lang.proxy_init(globals(), '.persistent.treapmap', [
        'TreapMap',
        'new_treap_map',
    ])

from .ranked import (  # noqa
    RankedSeq,
    RankedSetSeq,
)

if _ta.TYPE_CHECKING:
    from .sorted.skiplist import (  # noqa
        SkipList,
        SkipListDict,
    )
else:
    _lang.proxy_init(globals(), '.sorted.skiplist', [
        'SkipList',
        'SkipListDict',
    ])

from .sorted.sorted import (  # noqa
    SortedCollection,
    SortedItems,
    SortedIter,
    SortedListDict,
    SortedMapping,
    SortedMutableMapping,
)

from .unmodifiable import (  # noqa
    Unmodifiable,
    UnmodifiableMapping,
    UnmodifiableSequence,
    UnmodifiableSet,
)

from .utils import (  # noqa
    PartitionResult,
    all_equal,
    all_not_equal,
    indexes,
    key_cmp,
    make_map,
    make_map_by,
    multi_map,
    multi_map_by,
    partition,
    unique,
)



from ._registry import register


@register
class SpacingPlugin:
    def pytest_addoption(self, parser):
        parser.addoption('--newlines-before', action='store', type=int, help='Adds newlines before tests')
        parser.addoption('--newlines-after', action='store', type=int, help='Adds newlines after tests')

    def pytest_runtest_setup(self, item):
        if item.session.config.option.newlines_before:
            for _ in range(item.session.config.option.newlines_before):
                print()

    def pytest_runtest_teardown(self, item):
        if item.session.config.option.newlines_after:
            for _ in range(item.session.config.option.newlines_after):
                print()



# ruff: noqa: UP006 UP007
# @omlish-lite
import dataclasses as dc
import logging
import typing as ta

from ..addresses import SocketAndAddress
from ..io import close_socket_immediately
from .handlers import SocketServerHandler
from .handlers import SocketServerHandler_


##


@dc.dataclass(frozen=True)
class SslErrorHandlingSocketServerHandler(SocketServerHandler_):
    handler: SocketServerHandler

    log: ta.Optional[logging.Logger] = None

    #

    _error_cls: ta.ClassVar[ta.Optional[ta.Type[BaseException]]] = None

    @classmethod
    def _get_error_cls(cls) -> ta.Type[BaseException]:
        if (error_cls := cls._error_cls) is None:
            import ssl
            error_cls = cls._error_cls = ssl.SSLError
        return error_cls

    def __call__(self, conn: SocketAndAddress) -> None:
        try:
            self.handler(conn)
        except self._get_error_cls():  # noqa
            if (log := self.log) is not None:
                log.exception('SSL Error in connection %r', conn)
            close_socket_immediately(conn.socket)



import dataclasses as dc
import typing as ta

from .... import lang
from .building import JsonObjectBuilder
from .lexing import JsonStreamLexer
from .parsing import JsonStreamParser


##


@dc.dataclass(kw_only=True)
class JsonStreamObjectParser(lang.ExitStacked):
    include_raw: bool = False
    yield_object_lists: bool = False

    #

    _lex: JsonStreamLexer = dc.field(init=False)
    _parse: JsonStreamParser = dc.field(init=False)
    _build: JsonObjectBuilder = dc.field(init=False)

    def _enter_contexts(self) -> None:
        self._lex = JsonStreamLexer(
            include_raw=self.include_raw,
        )

        self._parse = JsonStreamParser()

        self._build = JsonObjectBuilder(
            yield_object_lists=self.yield_object_lists,
        )

    def feed(self, i: ta.Iterable[str]) -> ta.Iterator[ta.Any]:
        for c in i:
            for t in self._lex(c):
                for e in self._parse(t):
                    for v in self._build(e):  # noqa
                        yield v


def stream_parse_one_object(
        i: ta.Iterable[str],
        **kwargs: ta.Any,
) -> ta.Any:
    with JsonStreamObjectParser(**kwargs) as p:
        return next(p.feed(i))



from .base import (  # noqa
    DEFAULT_ENCODING,

    is_success_status,

    HttpRequest,

    BaseHttpResponse,
    HttpResponse,
    StreamHttpResponse,

    close_response,
    closing_response,
    read_response,

    HttpClientError,
    HttpStatusError,

    HttpClient,
)

from .default import (  # noqa
    client,

    request,
)

from .httpx import (  # noqa
    HttpxHttpClient,
)

from .urllib import (  # noqa
    UrllibHttpClient,
)



import abc
import functools
import typing as ta

from ... import cached
from ... import check
from ... import dataclasses as dc
from ... import lang
from ..qualifiedname import QualifiedName
from .base import HasQn
from .base import Node
from .idents import CanIdent
from .idents import Ident
from .idents import IdentBuilder
from .idents import as_ident


##


class NameLike(abc.ABC):  # noqa
    pass


##


def _coerce_name_parts(o: ta.Iterable[Ident]) -> ta.Sequence[Ident]:
    check.not_isinstance(o, str)
    return check.not_empty(tuple(check.isinstance(e, Ident) for e in o))


class Name(Node, NameLike, HasQn, lang.Final):
    ps: ta.Sequence[Ident] = dc.xfield(coerce=_coerce_name_parts)

    @cached.property
    def qn(self) -> QualifiedName:
        return QualifiedName(tuple(p.s for p in self.ps))


##


@functools.singledispatch
def as_name(o: ta.Any) -> Name:
    if isinstance(o, ta.Sequence):
        return Name([as_ident(p) for p in o])
    else:
        raise TypeError(o)


@as_name.register
def _(n: Name) -> Name:
    return n


@as_name.register
def _(o: Ident | str) -> Name:
    return Name([as_ident(o)])


##


CanName: ta.TypeAlias = ta.Union[
    NameLike,
    Name,
    CanIdent,
    ta.Sequence[CanIdent],
    'NameAccessor',
]


class NameAccessor(NameLike, lang.Final):
    def __init__(self, ps: tuple[str, ...] = ()) -> None:
        super().__init__()
        self.__query_name_parts__ = ps

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.__query_name_parts__!r})'

    def __getattr__(self, s: str) -> 'NameAccessor':
        return NameAccessor((*self.__query_name_parts__, s))

    def __call__(self, o: CanName) -> Name:
        n = as_name(o)
        if (ps := self.__query_name_parts__):
            n = Name(tuple(*tuple(Ident(p) for p in ps), *n.ps))  # type: ignore[arg-type]
        return n


@as_name.register
def _(a: NameAccessor) -> Name:
    return Name(tuple(Ident(p) for p in a.__query_name_parts__))


##


class NameBuilder(IdentBuilder):
    @ta.final
    def name(self, o: CanName) -> Name:
        return as_name(o)

    @ta.final
    @property
    def n(self) -> NameAccessor:
        return NameAccessor()



import typing as ta

from .. import lang
from .base import MarshalContext
from .base import MarshalerFactory
from .base import UnmarshalContext
from .base import UnmarshalerFactory
from .registries import Registry
from .registries import RegistryItem
from .standard import new_standard_marshaler_factory
from .standard import new_standard_unmarshaler_factory
from .values import Value


T = ta.TypeVar('T')


##


GLOBAL_REGISTRY: Registry = Registry()


##


@lang.cached_function(lock=True)
def global_marshaler_factory() -> MarshalerFactory:
    return new_standard_marshaler_factory()


def marshal(obj: ta.Any, ty: type | None = None, **kwargs: ta.Any) -> Value:
    mc = MarshalContext(GLOBAL_REGISTRY, factory=global_marshaler_factory(), **kwargs)
    return mc.make(ty if ty is not None else type(obj)).marshal(mc, obj)


##


@lang.cached_function(lock=True)
def global_unmarshaler_factory() -> UnmarshalerFactory:
    return new_standard_unmarshaler_factory()


def unmarshal(v: Value, ty: type[T], **kwargs: ta.Any) -> T:
    uc = UnmarshalContext(GLOBAL_REGISTRY, factory=global_unmarshaler_factory(), **kwargs)
    return uc.make(ty).unmarshal(uc, v)


##


def register_global(
        key: ta.Any,
        *items: RegistryItem,
        identity: bool = False,
) -> None:
    GLOBAL_REGISTRY.register(
        key,
        *items,
        identity=identity,
    )



from .attrs import (  # noqa
    AttrOps,
    AttributePresentError,
    DictAttrOps,
    STD_ATTR_OPS,
    SetAttrIfPresent,
    StdAttrOps,
    TRANSIENT_ATTR_OPS,
    TransientAttrOps,
    TransientDict,
    set_attr,
    transient_delattr,
    transient_getattr,
    transient_setattr,
)

from .cached.function import (  # noqa
    cached_function,
    static_init,
)

from .cached.property import (  # noqa
    cached_property,
)

from .casing import (  # noqa
    StringCasingError,
    ImproperStringCasingError,
    UnknownStringCasingError,
    AmbiguousStringCasingError,

    StringCasing,
    CamelCase,
    LowCamelCase,
    SnakeCase,
    UpSnakeCase,

    STRING_CASINGS,
    CAMEL_CASE,
    LOW_CAMEL_CASE,
    SNAKE_CASE,
    UP_SNAKE_CASE,

    camel_case,
    low_camel_case,
    snake_case,
    up_snake_case,

    get_string_casing,
    split_string_casing,
)

from .classes.abstract import (  # noqa
    Abstract,
    AbstractTypeError,
    get_abstract_methods,
    is_abstract,
    is_abstract_class,
    is_abstract_method,
    make_abstract,
    unabstract_class,
)

from .classes.restrict import (  # noqa
    AnySensitive,
    Final,
    FinalTypeError,
    NoBool,
    NotInstantiable,
    NotPicklable,
    PackageSealed,
    SENSITIVE_ATTR,
    Sealed,
    SealedError,
    Sensitive,
    no_bool,
)

from .classes.simple import (  # noqa
    LazySingleton,
    Marker,
    Namespace,
    SimpleMetaDict,
    Singleton,
)

from .classes.virtual import (  # noqa
    Callable,
    Descriptor,
    Picklable,
    Virtual,
    virtual_check,
)

from .clsdct import (  # noqa
    ClassDctFn,
    cls_dct_fn,
    get_caller_cls_dct,
    is_possibly_cls_dct,
)

from .collections import (  # noqa
    empty_map,
    merge_dicts,
    yield_dict_init,
)

from .comparison import (  # noqa
    Infinity,
    InfinityType,
    NegativeInfinity,
    NegativeInfinityType,
    cmp,
)

from .contextmanagers import (  # noqa
    AsyncContextManager,
    ContextManaged,
    ContextManager,
    ContextWrapped,
    DefaultLockable,
    Lockable,
    NOP_CONTEXT_MANAGER,
    NopContextManager,
    Timer,
    breakpoint_on_exception,
    context_var_setting,
    context_wrapped,
    default_lock,
    disposing,
    maybe_managing,
)

from .datetimes import (  # noqa
    utcnow,
    utcfromtimestamp,
)

from .descriptors import (  # noqa
    AccessForbiddenError,
    access_forbidden,
    attr_property,
    classonly,
    decorator,
    is_method_descriptor,
    item_property,
    unwrap_func,
    unwrap_func_with_partials,
    unwrap_method_descriptors,
    update_wrapper,
)

from .enums import (  # noqa
    enum_name_repr,
)

from .functions import (  # noqa
    Args,
    VoidError,
    as_async,
    call_with,
    coalesce,
    constant,
    finally_,
    identity,
    is_lambda,
    is_none,
    is_not_none,
    isinstance_of,
    issubclass_of,
    maybe_call,
    new_function,
    new_function_kwargs,
    opt_coalesce,
    opt_fn,
    opt_kw,
    periodically,
    raise_,
    raising,
    recurse,
    strict_constant,
    try_,
    void,
)

from .generators import (  # noqa
    GeneratorLike,
    GeneratorMappedIterator,
    autostart,
    capture_coroutine,
    capture_generator,
    genmap,
    nextgen,
)

from .imports import (  # noqa
    LazyGlobals,
    can_import,
    get_real_module_name,
    import_all,
    lazy_import,
    proxy_import,
    proxy_init,
    resolve_import_name,
    try_import,
    yield_import_all,
    yield_importable,
)

from .iterables import (  # noqa
    asrange,
    common_prefix_len,
    exhaust,
    flatmap,
    flatten,
    ilen,
    interleave,
    itergen,
    peek,
    prodrange,
    readiter,
    renumerate,
    take,
)

from .maybes import (  # noqa
    Maybe,
    empty,
    just,
    maybe,
)

from .objects import (  # noqa
    Identity,
    SimpleProxy,
    anon_object,
    arg_repr,
    can_weakref,
    deep_subclasses,
    dir_dict,
    mro_dict,
    mro_owner_dict,
    new_type,
    opt_repr,
    super_meta,
)

from .outcomes import (  # noqa
    Either,
    Error,
    Outcome,
    OutcomeAlreadyUnwrappedError,
    Value,
    acapture,
    capture,
    error,
    value,
)

from .params import (  # noqa
    ArgsParam,
    KwOnlyParam,
    KwargsParam,
    Param,
    ParamSeparator,
    ParamSpec,
    PosOnlyParam,
    ValParam,
    VarParam,
    param_render,
)

from .recursion import (  # noqa
    LimitedRecursionError,
    recursion_limiting,
    recursion_limiting_context,
)

from .resolving import (  # noqa
    Resolvable,
    ResolvableClassNameError,
    get_cls_fqcn,
    get_fqcn_cls,
)

from .resources import (  # noqa
    ReadableResource,
    get_package_resources,
    get_relative_resources,
)

from .strings import (  # noqa
    BOOL_FALSE_STRINGS,
    BOOL_STRINGS,
    BOOL_TRUE_STRINGS,
    STRING_BOOL_VALUES,
    find_any,
    indent_lines,
    is_dunder,
    is_ident,
    is_ident_cont,
    is_ident_start,
    is_sunder,
    prefix_delimited,
    prefix_lines,
    replace_many,
    rfind_any,
    strip_prefix,
    strip_suffix,
)

from .sys import (  # noqa
    REQUIRED_PYTHON_VERSION,
    check_runtime_version,
    is_gil_enabled,
)

from .typing import (  # noqa
    BytesLike,
    SequenceNotStr,
    protocol_check,
    typed_lambda,
    typed_partial,
)

##

from ..lite.contextmanagers import (  # noqa
    AsyncExitStacked,
    ExitStacked,
    adefer,
    attr_setting,
    defer,
)

from ..lite.imports import (  # noqa
    import_attr,
    import_module,
    import_module_attr,
)

from ..lite.reprs import (  # noqa
    AttrRepr,
    attr_repr,
)

from ..lite.timeouts import (  # noqa
    DeadlineTimeout,
    InfiniteTimeout,
    Timeout,
    TimeoutLike,
)

from ..lite.types import (  # noqa
    BUILTIN_SCALAR_ITERABLE_TYPES,
)

from ..lite.typing import (  # noqa
    AnyFunc,
    Func0,
    Func1,
    Func2,
    Func3,
)



# ruff: noqa: UP006 UP007
# @omlish-lite
"""
TODO:
 - raw
 - blocks / inheritance
"""
import dataclasses as dc
import io
import re
import typing as ta

from ..lite.cached import cached_nullary
from ..lite.check import check
from ..lite.maybes import Maybe


##


@dc.dataclass(frozen=True)
class MinjaTemplateParam:
    name: str
    default: Maybe[ta.Any] = Maybe.empty()

    def __post_init__(self) -> None:
        check.arg(self.name.isidentifier())

    @classmethod
    def of(cls, obj: ta.Union[str, 'MinjaTemplateParam']) -> 'MinjaTemplateParam':
        if isinstance(obj, MinjaTemplateParam):
            return obj
        elif isinstance(obj, str):
            return MinjaTemplateParam.new(obj)
        else:
            raise TypeError(obj)

    @classmethod
    def new(cls, name: str, *defaults: ta.Any) -> 'MinjaTemplateParam':
        dfl: Maybe[ta.Any]
        if defaults:
            [dv] = defaults
            dfl = Maybe.just(dv)
        else:
            dfl = Maybe.empty()
        return cls(name, dfl)


class MinjaTemplate:
    def __init__(
            self,
            fn: ta.Callable,
            params: ta.Sequence[MinjaTemplateParam],
    ) -> None:
        super().__init__()

        self._fn = fn
        self._params = params

    @property
    def params(self) -> ta.Sequence[MinjaTemplateParam]:
        return self._params

    def __call__(self, **kwargs: ta.Any) -> str:
        return self._fn(**kwargs)


##


class MinjaTemplateCompiler:
    """
    Compiles a template string into a Python function. The returned function takes a dictionary 'context' and returns
    the rendered string.

    Supported syntax:
    - Literal text remains literal.
    - {{ expr }}: Evaluates 'expr' in the given context and writes its str() to output.
    - {% code %}: Executes the given Python code (e.g. 'for x in items:'), must be terminated appropriately.
    - {% endfor %} to close for loops.
    - {% endif %} to close if blocks.
    - {# comment #}: Ignored completely.
    """

    DEFAULT_INDENT: str = ' ' * 4

    def __init__(
            self,
            src: str,
            params: ta.Sequence[ta.Union[str, MinjaTemplateParam]],
            *,
            indent: str = DEFAULT_INDENT,
    ) -> None:
        super().__init__()

        check.not_isinstance(params, str)

        self._src = check.isinstance(src, str)
        self._params = [
            MinjaTemplateParam.of(p)
            for p in params
        ]
        check.unique(p.name for p in self._params)
        self._indent_str: str = check.non_empty_str(indent)

        self._stack: ta.List[ta.Literal['for', 'if']] = []

    #

    _TAG_PAT = re.compile(
        r'({{.*?}}|{%.*?%}|{#.*?#})',
        flags=re.DOTALL,
    )

    @classmethod
    def _split_tags(cls, src: str) -> ta.List[ta.Tuple[str, str]]:
        raw = cls._TAG_PAT.split(src)

        #

        parts: ta.List[ta.Tuple[str, str]] = []
        for s in raw:
            if not s:
                continue

            for g, l, r in [
                ('{', '{{', '}}'),
                ('%', '{%', '%}'),
                ('#', '{#', '#}'),
            ]:
                if s.startswith(l) and s.endswith(r):
                    parts.append((g, s[len(l):-len(r)]))
                    break
            else:
                parts.append(('', s))

        #

        for i, (g, s) in enumerate(parts):
            if s.startswith('-'):
                if i > 0:
                    lg, ls = parts[i - 1]
                    parts[i - 1] = (lg, ls.rstrip())
                s = s[1:].lstrip()

            if s.endswith('-'):
                if i < len(parts) - 1:
                    rg, rs = parts[i + 1]
                    parts[i + 1] = (rg, rs.lstrip())
                s = s[:-1].rstrip()

            parts[i] = (g, s)

        #

        parts = [(g, s) for g, s in parts if g or s]

        #

        return parts

    #

    def _indent(self, line: str, ofs: int = 0) -> str:
        return self._indent_str * (len(self._stack) + 1 + ofs) + line

    #

    _RENDER_FN_NAME = '__render'

    class Rendered(ta.NamedTuple):
        src: str
        ns: ta.Dict[str, ta.Any]

    @cached_nullary
    def render(self) -> Rendered:
        lines: ta.List[str] = []

        ns: ta.Dict[str, ta.Any] = {
            '__StringIO': io.StringIO,
        }

        parts = self._split_tags(self._src)

        if not self._params:
            lines.append(f'def {self._RENDER_FN_NAME}():')
        else:
            lines.append(f'def {self._RENDER_FN_NAME}(')
            for p in self._params:
                if p.default.present:
                    check.not_in(p.name, ns)
                    ns[p.name] = p.default.must()
                    lines.append(self._indent(f'{p.name}={p.name},'))
                else:
                    lines.append(self._indent(f'{p.name},'))
            lines.append('):')

        lines.append(self._indent('__output = __StringIO()'))

        for g, s in parts:
            if g == '{':
                expr = s.strip()
                lines.append(self._indent(f'__output.write(str({expr}))'))

            elif g == '%':
                stmt = s.strip()

                if stmt.startswith('for '):
                    lines.append(self._indent(stmt + ':'))
                    self._stack.append('for')
                elif stmt.startswith('endfor'):
                    check.equal(self._stack.pop(), 'for')

                elif stmt.startswith('if '):
                    lines.append(self._indent(stmt + ':'))
                    self._stack.append('if')
                elif stmt.startswith('elif '):
                    check.equal(self._stack[-1], 'if')
                    lines.append(self._indent(stmt + ':', -1))
                elif stmt.strip() == 'else':
                    check.equal(self._stack[-1], 'if')
                    lines.append(self._indent('else:', -1))
                elif stmt.startswith('endif'):
                    check.equal(self._stack.pop(), 'if')

                else:
                    lines.append(self._indent(stmt))

            elif g == '#':
                pass

            elif not g:
                if s:
                    safe_text = s.replace('"""', '\\"""')
                    lines.append(self._indent(f'__output.write("""{safe_text}""")'))

            else:
                raise KeyError(g)

        check.empty(self._stack)

        lines.append(self._indent('return __output.getvalue()'))

        return self.Rendered('\n'.join(lines), ns)

    #

    @classmethod
    def _make_fn(
            cls,
            name: str,
            src: str,
            ns: ta.Optional[ta.Mapping[str, ta.Any]] = None,
    ) -> ta.Callable:
        glo: dict = {}
        if ns:
            glo.update(ns)
        exec(src, glo)
        return glo[name]

    #

    @cached_nullary
    def compile(self) -> MinjaTemplate:
        rendered = self.render()

        render_fn = self._make_fn(
            self._RENDER_FN_NAME,
            rendered.src,
            rendered.ns,
        )

        return MinjaTemplate(
            render_fn,
            self._params,
        )


##


def compile_minja_template(
        src: str,
        params: ta.Sequence[ta.Union[str, MinjaTemplateParam]] = (),
) -> MinjaTemplate:
    return MinjaTemplateCompiler(src, params).compile()


def render_minja_template(src: str, **kwargs: ta.Any) -> str:
    tmpl = compile_minja_template(src, list(kwargs))
    return tmpl(**kwargs)



"""
TODO:
 - find_vars
"""
import abc
import dataclasses as dc
import string
import typing as ta

from .. import lang
from .minja import MinjaTemplate
from .minja import MinjaTemplateParam
from .minja import compile_minja_template


if ta.TYPE_CHECKING:
    import jinja2
else:
    jinja2 = lang.proxy_import('jinja2')


##


class Templater(lang.Abstract):
    @dc.dataclass(frozen=True)
    class Context:
        env: ta.Mapping[str, ta.Any] | None = None

    @abc.abstractmethod
    def render(self, ctx: Context) -> str:
        raise NotImplementedError


##


@dc.dataclass(frozen=True)
class FormatTemplater(Templater):
    """https://docs.python.org/3/library/string.html#format-specification-mini-language"""

    fmt: str

    def render(self, ctx: Templater.Context) -> str:
        return self.fmt.format(**(ctx.env or {}))


##


@dc.dataclass(frozen=True)
class Pep292Templater(Templater):
    """https://peps.python.org/pep-0292/"""

    tmpl: string.Template

    def render(self, ctx: Templater.Context) -> str:
        return self.tmpl.substitute(ctx.env or {})

    @classmethod
    def from_string(cls, src: str) -> 'Pep292Templater':
        return cls(string.Template(src))

##


@dc.dataclass(frozen=True)
class MinjaTemplater(Templater):
    tmpl: MinjaTemplate

    ENV_IDENT: ta.ClassVar[str] = 'env'

    def render(self, ctx: Templater.Context) -> str:
        return self.tmpl(**{self.ENV_IDENT: ctx.env or {}})

    @classmethod
    def from_string(
            cls,
            src: str,
            **ns: ta.Any,
    ) -> 'MinjaTemplater':
        tmpl = compile_minja_template(
            src,
            [
                MinjaTemplateParam(cls.ENV_IDENT),
                *[
                    MinjaTemplateParam.new(k, v)
                    for k, v in ns.items()
                ],
            ],
        )

        return cls(tmpl)


##


@dc.dataclass(frozen=True)
class JinjaTemplater(Templater):
    tmpl: 'jinja2.Template'

    def render(self, ctx: Templater.Context) -> str:
        return self.tmpl.render(**(ctx.env or {}))

    @classmethod
    def from_string(
            cls,
            src: str,
            *,
            env: ta.Optional['jinja2.Environment'] = None,
            **kwargs: ta.Any,
    ) -> 'JinjaTemplater':
        if env is None:
            env = jinja2.Environment()
        tmpl = env.from_string(src)
        return cls(
            tmpl,
            **kwargs,
        )



import types
import typing as ta
import weakref

from .classes.abstract import is_abstract_class


T = ta.TypeVar('T')


##


def arg_repr(*args: ta.Any, **kwargs: ta.Any) -> str:
    return ', '.join(*(
        list(map(repr, args)) +
        [f'{k}={v!r}' for k, v in kwargs.items()]
    ))


def opt_repr(obj: ta.Any) -> str | None:
    if obj is None:
        return None
    return repr(obj)


##


_CAN_WEAKREF_TYPE_MAP: ta.MutableMapping[type, bool] = weakref.WeakKeyDictionary()


def can_weakref(obj: ta.Any) -> bool:
    _type = type(obj)
    try:
        return _CAN_WEAKREF_TYPE_MAP[_type]
    except KeyError:
        pass
    try:
        weakref.ref(obj)
    except TypeError:
        ret = False
    else:
        ret = True
    _CAN_WEAKREF_TYPE_MAP[_type] = ret
    return ret


##


def new_type(
        name: str,
        bases: ta.Sequence[ta.Any],
        namespace: ta.Mapping[str, ta.Any],
        **kwargs: ta.Any,
) -> type:
    return types.new_class(
        name,
        tuple(bases),
        kwds=kwargs,
        exec_body=lambda ns: ns.update(namespace),
    )


def super_meta(
        super_meta: ta.Any,
        meta: type,
        name: str,
        bases: ta.Sequence[ta.Any],
        namespace: ta.MutableMapping[str, ta.Any],
        **kwargs: ta.Any,
) -> type:
    """Per types.new_class"""
    resolved_bases = types.resolve_bases(bases)
    if resolved_bases is not bases:
        if '__orig_bases__' in namespace:
            raise TypeError((bases, resolved_bases))
        namespace['__orig_bases__'] = bases
    return super_meta.__new__(meta, name, resolved_bases, dict(namespace), **kwargs)


##


def deep_subclasses(
        cls: type[T],
        *,
        concrete_only: bool = False,
) -> ta.Iterator[type[T]]:
    seen = set()
    todo = list(reversed(cls.__subclasses__()))
    while todo:
        cur = todo.pop()
        if cur in seen:
            continue
        seen.add(cur)
        if not (concrete_only and is_abstract_class(cur)):
            yield cur
        todo.extend(reversed(cur.__subclasses__()))


def mro_owner_dict(
        instance_cls: type,
        owner_cls: type | None = None,
        *,
        bottom_up_key_order: bool = False,
        sort_keys: bool = False,
) -> ta.Mapping[str, tuple[type, ta.Any]]:
    if owner_cls is None:
        owner_cls = instance_cls

    mro = instance_cls.__mro__[-2::-1]
    try:
        pos = mro.index(owner_cls)
    except ValueError:
        raise TypeError(f'Owner class {owner_cls} not in mro of instance class {instance_cls}') from None

    dct: dict[str, tuple[type, ta.Any]] = {}
    if not bottom_up_key_order:
        for cur_cls in mro[:pos + 1][::-1]:
            for k, v in cur_cls.__dict__.items():
                if k not in dct:
                    dct[k] = (cur_cls, v)

    else:
        for cur_cls in mro[:pos + 1]:
            dct.update({k: (cur_cls, v) for k, v in cur_cls.__dict__.items()})

    if sort_keys:
        dct = dict(sorted(dct.items(), key=lambda t: t[0]))

    return dct


def mro_dict(
        instance_cls: type,
        owner_cls: type | None = None,
        *,
        bottom_up_key_order: bool = False,
        sort_keys: bool = False,
) -> ta.Mapping[str, ta.Any]:
    return {
        k: v
        for k, (o, v) in mro_owner_dict(
            instance_cls,
            owner_cls,
            bottom_up_key_order=bottom_up_key_order,
            sort_keys=sort_keys,
        ).items()
    }


def dir_dict(o: ta.Any) -> dict[str, ta.Any]:
    return {
        a: getattr(o, a)
        for a in dir(o)
    }


##


class SimpleProxy(ta.Generic[T]):
    class Descriptor:
        def __init__(self, attr: str) -> None:
            super().__init__()
            self._attr = attr

        def __get__(self, instance, owner=None):
            if instance is None:
                return self
            return getattr(object.__getattribute__(instance, '__wrapped__'), self._attr)

        def __set__(self, instance, value):
            if instance is None:
                return self
            setattr(object.__getattribute__(instance, '__wrapped__'), self._attr, value)
            return None

        def __delete__(self, instance):
            if instance is None:
                return self
            delattr(object.__getattribute__(instance, '__wrapped__'), self._attr)
            return None

    __wrapped_attrs__: ta.ClassVar[ta.Iterable[str]] = ()

    def __init__(self, wrapped: T) -> None:
        super().__init__()
        object.__setattr__(self, '__wrapped__', wrapped)

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        for attr in cls.__wrapped_attrs__:
            setattr(cls, attr, SimpleProxy.Descriptor(attr))

    def __getattr__(self, item):
        return getattr(object.__getattribute__(self, '__wrapped__'), item)

    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, '__wrapped__'), name, value)

    def __delattr__(self, item):
        delattr(object.__getattribute__(self, '__wrapped__'), item)


##


class _AnonObject:
    def __init__(self, **attrs: ta.Any) -> None:
        super().__init__()

        for k, v in attrs.items():
            setattr(self, k, v)

    def __init_subclass__(cls, **kwargs):
        raise TypeError


def anon_object(**attrs: ta.Any) -> ta.Any:
    return _AnonObject(**attrs)


##


class Identity(ta.Generic[T]):
    def __init__(self, obj: T) -> None:
        super().__init__()

        self._obj = obj

    def __bool__(self):
        raise TypeError

    @property
    def obj(self) -> T:
        return self._obj

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._obj!r})'

    def __hash__(self) -> int:
        return id(self._obj)

    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        return self._obj is other._obj  # noqa



import operator as op
import typing as ta
import weakref

from .. import lang


T = ta.TypeVar('T')
K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class IdentityWrapper(ta.Generic[T]):
    def __init__(self, value: T) -> None:
        super().__init__()

        self._value = value

    def __repr__(self) -> str:
        return lang.attr_repr(self, 'value')

    @property
    def value(self) -> T:
        return self._value

    def __eq__(self, other):
        return isinstance(other, IdentityWrapper) and other._value is self._value

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return id(self._value)


class IdentityKeyDict(ta.MutableMapping[K, V]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

        self._dict: dict[int, tuple[K, V]] = {}
        for k, v in lang.yield_dict_init(*args, **kwargs):
            self[k] = v

    def __reduce__(self):
        return (type(self), (list(self.items()),))

    @property
    def debug(self) -> ta.Sequence[tuple[K, V]]:
        return list(self.items())

    def __repr__(self) -> str:
        return lang.attr_repr(self, '_dict')

    def __setitem__(self, k: K, v: V) -> None:
        self._dict[id(k)] = (k, v)

    def __delitem__(self, k: K) -> None:
        del self._dict[id(k)]

    def __getitem__(self, k: K) -> V:
        return self._dict[id(k)][1]

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> ta.Iterator[K]:
        return iter(map(op.itemgetter(0), self._dict.values()))

    def clear(self) -> None:
        self._dict.clear()


class IdentitySet(ta.MutableSet[T]):
    def __init__(self, init: ta.Iterable[T] | None = None) -> None:
        super().__init__()

        self._dict: dict[int, T] = {}
        if init is not None:
            for item in init:
                self.add(item)

    def __reduce__(self):
        return (type(self), (list(self),))

    @property
    def debug(self) -> ta.Sequence[T]:
        return list(self)

    def __repr__(self) -> str:
        return lang.attr_repr(self, '_dict')

    def add(self, item: T) -> None:
        self._dict[id(item)] = item

    def discard(self, item: T) -> None:
        try:
            del self._dict[id(item)]
        except KeyError:
            pass

    def update(self, items: ta.Iterable[T]) -> None:
        for item in items:
            self.add(item)

    def __contains__(self, item: T) -> bool:  # type: ignore
        return id(item) in self._dict

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> ta.Iterator[T]:
        return iter(self._dict.values())


class IdentityWeakSet(ta.MutableSet[T]):
    def __init__(self, init: ta.Iterable[T] | None = None) -> None:
        super().__init__()

        self._dict: weakref.WeakValueDictionary[int, T] = weakref.WeakValueDictionary()

    def add(self, value):
        # FIXME: race with weakref callback?
        self._dict[id(value)] = value

    def discard(self, value):
        try:
            del self._dict[id(value)]
        except KeyError:
            pass

    def __contains__(self, x):
        try:
            o = self._dict[id(x)]
        except KeyError:
            return False
        return x is o

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return self._dict.values()



import os
import sys


def reparent_process(
        *,
        no_close_stdio: bool = False,
) -> None:
    if (pid := os.fork()):  # noqa
        sys.exit(0)
        raise RuntimeError('Unreachable')  # noqa

    os.setsid()

    if (pid := os.fork()):  # noqa
        sys.exit(0)

    if not no_close_stdio:
        rn_fd = os.open('/dev/null', os.O_RDONLY)
        os.dup2(rn_fd, 0)
        os.close(rn_fd)

        wn_fd = os.open('/dev/null', os.O_WRONLY)
        os.dup2(wn_fd, 1)
        os.dup2(wn_fd, 2)
        os.close(wn_fd)

    sys.stdout.flush()
    sys.stderr.flush()



import shutil
import sys
import sysconfig
import typing as ta

import pytest

from ... import lang


##


def if_cant_import(module: str, *args, **kwargs):
    return pytest.mark.skipif(not lang.can_import(module, *args, **kwargs), reason=f'requires import {module}')


def if_not_on_path(exe: str):
    return pytest.mark.skipif(shutil.which(exe) is None, reason=f'requires exe on path {exe}')


def if_python_version_less_than(num: ta.Sequence[int]):
    return pytest.mark.skipif(sys.version_info < tuple(num), reason=f'python version {tuple(sys.version_info)} < {tuple(num)}')  # noqa


def if_not_platform(*platforms: str):
    return pytest.mark.skipif(sys.platform not in platforms, reason=f'requires platform in {platforms}')


def if_not_single():
    # FIXME
    # [resolve_collection_argument(a) for a in session.config.args]
    raise NotImplementedError


def if_nogil():
    return pytest.mark.skipif(sysconfig.get_config_var('Py_GIL_DISABLED'), reason='requires gil build')



"""Should be kept pure. No references to dc std, no references to impl detail."""
import dataclasses as dc
import enum
import typing as ta

from .. import check
from .. import lang


T = ta.TypeVar('T')


##


CoerceFn: ta.TypeAlias = ta.Callable[[ta.Any], ta.Any]
ValidateFn: ta.TypeAlias = ta.Callable[[ta.Any], bool]
ReprFn: ta.TypeAlias = ta.Callable[[ta.Any], str | None]

InitFn: ta.TypeAlias = ta.Callable[[ta.Any], None]
ClassValidateFn: ta.TypeAlias = ta.Callable[..., bool]


class DefaultFactory(ta.NamedTuple):
    fn: ta.Callable[..., ta.Any]


##


class FieldType(enum.StrEnum):
    INSTANCE = enum.auto()
    CLASS_VAR = enum.auto()
    INIT_VAR = enum.auto()

    __repr__ = lang.enum_name_repr


@dc.dataclass(frozen=True, kw_only=True, eq=False)
class FieldSpec(lang.Final):
    name: str
    annotation: ta.Any

    default: lang.Maybe[DefaultFactory | ta.Any] = lang.empty()

    ##
    # std

    init: bool = True
    repr: bool = True

    # This can be a bool or None. If true, this field is included in the generated __hash__() method. If false, this
    # field is excluded from the generated __hash__(). If None (the default), use the value of compare: this would
    # normally be the expected behavior, since a field should be included in the hash if it's used for comparisons.
    # Setting this value to anything other than None is discouraged.
    hash: bool | None = None  # FIXME: type?

    compare: bool = True

    # This can be a mapping or None. None is treated as an empty dict. This value is wrapped in MappingProxyType() to
    # make it read-only, and exposed on the Field object. It is not used at all by Data Classes, and is provided as a
    # third-party extension mechanism. Multiple third-parties can each have their own key, to use as a namespace in the
    # metadata.
    metadata: ta.Mapping[ta.Any, ta.Any] | None = None

    kw_only: bool | None = None

    # doc: ta.Any = None

    ##
    # ext

    # derive: ta.Callable[..., ta.Any] | None = None  # NYI in core

    coerce: bool | CoerceFn | None = None
    validate: ValidateFn | None = None
    check_type: bool | type | tuple[type | None, ...] | None = None

    override: bool = False

    repr_fn: ReprFn | None = None
    repr_priority: int | None = None

    # frozen: bool | None = None  # NYI in core

    ##
    # derived

    field_type: FieldType = FieldType.INSTANCE

    ##
    # validate

    def __post_init__(self) -> None:
        check.non_empty_str(self.name)
        check.arg(self.name.isidentifier())

        if self.field_type in (FieldType.CLASS_VAR, FieldType.INIT_VAR):
            if isinstance(self.default.or_else(None), DefaultFactory):
                raise TypeError(f'field {self.name} cannot have a default factory')

        if self.field_type is FieldType.CLASS_VAR:
            if self.kw_only is not None:
                raise TypeError(f'field {self.name} is a ClassVar but specifies kw_only')
            check.none(self.coerce)
            check.none(self.validate)
            check.in_(self.check_type, (None, False))

        if (
                self.field_type is FieldType.INSTANCE and
                self.default.present and
                not isinstance(dfv := self.default.must(), DefaultFactory) and
                dfv.__class__.__hash__ is None  # noqa
        ):
            raise ValueError(
                f'mutable default {type(dfv)} for field {self.name} '
                f'is not allowed: use default_factory',
            )


##


@dc.dataclass(frozen=True, kw_only=True, eq=False)
class ClassSpec(lang.Final):
    ##
    # fields

    fields: ta.Sequence[FieldSpec]

    @property
    def fields_by_name(self) -> ta.Mapping[str, FieldSpec]:
        return self._fields_by_name

    _fields_by_name: ta.ClassVar[ta.Mapping[str, FieldSpec]]

    ##
    # std

    init: bool = True
    repr: bool = True
    eq: bool = True
    order: bool = False
    unsafe_hash: bool = False
    frozen: bool = False

    match_args: bool = True
    kw_only: bool = False
    slots: bool = False
    weakref_slot: bool = False

    ##
    # ext

    metadata: ta.Sequence[ta.Any] | None = None

    @property
    def metadata_by_type(self) -> ta.Mapping[type, ta.Sequence[ta.Any]]:
        return self._metadata_by_type

    _metadata_by_type: ta.ClassVar[ta.Mapping[type, ta.Sequence[ta.Any]]]

    @ta.overload
    def get_last_metadata(self, ty: type[T], default: T) -> T:
        ...

    @ta.overload
    def get_last_metadata(self, ty: type[T], default: None = None) -> T | None:
        ...

    def get_last_metadata(self, ty, default=None):
        try:
            mdl = self._metadata_by_type[ty]
        except KeyError:
            return default
        if not mdl:
            return default
        return mdl[-1]

    #

    reorder: bool = False
    cache_hash: bool = False
    generic_init: bool = False
    override: bool = False

    repr_id: bool = False
    terse_repr: bool = False

    allow_redundant_decorator: bool = False

    ##
    # callbacks

    init_fns: ta.Sequence[InitFn | property] | None = None

    @dc.dataclass(frozen=True)
    class ValidateFnWithParams:
        fn: ClassValidateFn
        params: ta.Sequence[str]

        def __post_init__(self) -> None:
            check.not_isinstance(self.params, str)

    validate_fns: ta.Sequence[ValidateFnWithParams] | None = None

    ##
    # validate

    def __post_init__(self) -> None:
        fields_by_name: dict[str, FieldSpec] = {}
        for f in self.fields:
            check.not_in(f.name, fields_by_name)
            fields_by_name[f.name] = f
        object.__setattr__(self, '_fields_by_name', fields_by_name)

        metadata_by_type: dict[type, list[ta.Any]] = {}
        for md in self.metadata or ():
            mdt = type(md)
            try:
                mdl = metadata_by_type[mdt]
            except KeyError:
                mdl = metadata_by_type[mdt] = []
            mdl.append(md)
        object.__setattr__(self, '_metadata_by_type', metadata_by_type)

        if self.order and not self.eq:
            raise ValueError('eq must be true if order is true')



# @omlish-manifest
_CLI_MODULE = {'$omdev.cli.types.CliModule': {
    'cmd_name': 'pidfiles',
    'mod_name': __name__,
}}


if __name__ == '__main__':
    from .cli import _main  # noqa

    _main()



# @omlish-lite
import urllib.request


##


class NonRaisingUrllibErrorProcessor(urllib.request.HTTPErrorProcessor):
    """
    https://stackoverflow.com/a/74844056

    Usage:

        opener = urllib.request.build_opener(NonRaisingUrllibErrorProcessor)
        with opener.open(req) as resp:
            ...
    """

    def http_response(self, request, response):
        return response

    def https_response(self, request, response):
        return response



from ...text.mangle import StringMangler


##


ESCAPED_IDENT_CHARS = ''.join(
    chr(i)
    for i in range(128)
    if not (
        chr(i).isalpha() or
        chr(i).isdigit() or
        chr(i) == '_' or
        ord(chr(i)) > 127
    )
)

IDENT_MANGLER = StringMangler.of('_', ESCAPED_IDENT_CHARS)



"""
TODO:
 - Namer: ta.TypeAlias = ta.Callable[[str], str] ?
  - this interface is ~intentionally~ limited, but custom overrides would be useful
"""
import enum

from .. import check
from .. import lang
from .base import Option


class Naming(Option, enum.Enum):
    SNAKE = 'snake'
    CAMEL = 'camel'
    LOW_CAMEL = 'low_camel'


def translate_name(n: str, e: Naming) -> str:
    check.non_empty_str(n)
    check.not_equal(set(n), {'_'})

    n1 = n.lstrip('_')
    pfx = '_' * (len(n) - len(n1))
    n2 = n1.rstrip('_')
    sfx = '_' * (len(n1) - len(n2))
    ps = lang.split_string_casing(n2)

    if e is Naming.SNAKE:
        r = lang.snake_case(*ps)
    elif e is Naming.CAMEL:
        r = lang.camel_case(*ps)
    elif e is Naming.LOW_CAMEL:
        r = lang.low_camel_case(*ps)
    else:
        raise ValueError(e)

    return pfx + r + sfx



import inspect
import typing as ta


def is_coroutine_function(func: ta.Any) -> bool:
    return inspect.iscoroutinefunction(func) or getattr(func, '_is_coroutine', False)


def is_async_function(func: ta.Any) -> bool:
    return is_coroutine_function(func) or inspect.isasyncgenfunction(func)



"""
TODO:
 - prebuild field frozenset for getters/setters
  - and one field per line
"""
import dataclasses as dc
import typing as ta
import weakref

from ... import check
from ..generation.base import Generator
from ..generation.base import Plan
from ..generation.base import PlanResult
from ..generation.globals import FROZEN_INSTANCE_ERROR_GLOBAL
from ..generation.idents import CLS_IDENT
from ..generation.idents import IDENT_PREFIX
from ..generation.ops import AddMethodOp
from ..generation.ops import Op
from ..generation.registry import register_generator_type
from ..internals import STD_FIELDS_ATTR
from ..internals import STD_PARAMS_ATTR
from ..processing.base import ProcessingContext


##


_UNCHECKED_FROZEN_BASES: ta.MutableSet[type] = weakref.WeakSet()


def unchecked_frozen_base(cls):
    _UNCHECKED_FROZEN_BASES.add(check.isinstance(cls, type))
    return cls


def check_frozen_bases(cls: type, frozen: bool) -> None:
    all_frozen_bases = None
    any_frozen_base = False
    has_dataclass_bases = False

    for b in cls.__mro__[-1:0:-1]:
        if b in _UNCHECKED_FROZEN_BASES:
            continue

        base_fields = getattr(b, STD_FIELDS_ATTR, None)
        if base_fields is None:
            continue

        has_dataclass_bases = True
        if all_frozen_bases is None:
            all_frozen_bases = True

        current_frozen = getattr(b, STD_PARAMS_ATTR).frozen
        all_frozen_bases = all_frozen_bases and current_frozen
        any_frozen_base = any_frozen_base or current_frozen

    if has_dataclass_bases:
        if any_frozen_base and not frozen:
            raise TypeError('cannot inherit non-frozen dataclass from a frozen one')

        if all_frozen_bases is False and frozen:
            raise TypeError('cannot inherit frozen dataclass from a non-frozen one')


##


@dc.dataclass(frozen=True)
class FrozenPlan(Plan):
    fields: tuple[str, ...]


@register_generator_type(FrozenPlan)
class FrozenGenerator(Generator[FrozenPlan]):
    def plan(self, ctx: ProcessingContext) -> PlanResult[FrozenPlan] | None:
        check_frozen_bases(ctx.cls, ctx.cs.frozen)

        if not ctx.cs.frozen:
            return None

        return PlanResult(FrozenPlan(tuple(f.name for f in ctx.cs.fields)))

    def _generate_one(
            self,
            plan: FrozenPlan,
            mth: str,
            params: ta.Sequence[str],
            exc_args: str,
    ) -> AddMethodOp:
        preamble = []
        # https://github.com/python/cpython/commit/ee6f8413a99d0ee4828e1c81911e203d3fff85d5
        condition = f'type(self) is {CLS_IDENT}'

        if plan.fields:
            set_ident = f'{IDENT_PREFIX}_{mth}_frozen_fields'
            preamble.extend([
                f'{set_ident} = {{',
                *[
                    f'    {f!r},'
                    for f in plan.fields
                ],
                f'}}',
                f'',
            ])
            condition += f' or name in {set_ident}'

        return AddMethodOp(
            f'__{mth}__',
            '\n'.join([
                *preamble,
                f'def __{mth}__(self, {", ".join(params)}):',
                f'    if {condition}:',
                f'        raise {FROZEN_INSTANCE_ERROR_GLOBAL.ident}{exc_args}',
                f'    super({CLS_IDENT}, self).__{mth}__({", ".join(params)})',
            ]),
            frozenset([FROZEN_INSTANCE_ERROR_GLOBAL]),
        )

    def generate(self, plan: FrozenPlan) -> ta.Iterable[Op]:
        return [
            self._generate_one(
                plan,
                'setattr',
                ['name', 'value'],
                '(f"cannot assign to field {name!r}")',
            ),
            self._generate_one(
                plan,
                'delattr',
                ['name'],
                '(f"cannot delete field {name!r}")',
            ),
        ]



import typing as ta

from .. import json
from ..codecs import make_object_lazy_loaded_codec
from ..codecs import make_str_object_codec
from .parsing import parse


##


def dumps(obj: ta.Any) -> str:
    return json.dumps(obj)


def loads(s: str) -> ta.Any:
    return parse(s)


##


JSON5_CODEC = make_str_object_codec('json5', dumps, loads)

# @omlish-manifest
_JSON5_LAZY_CODEC = make_object_lazy_loaded_codec(__name__, 'JSON5_CODEC', JSON5_CODEC)



"""
TODO:
 - visitor / transformer
 - uniform collection isinstance - items() for mappings, iter() for other
 - also check instance type in isinstance not just items lol
 - ta.Generic in mro causing trouble - omit? no longer 1:1
 - cache this shit, esp generic_mro shit
  - cache __hash__ in Generic/Union
"""
import abc
import dataclasses as dc
import types
import typing as ta


_NoneType = types.NoneType  # type: ignore

_NONE_TYPE_FROZENSET: frozenset['Type'] = frozenset([_NoneType])


_AnnotatedAlias = ta._AnnotatedAlias  # type: ignore  # noqa
_CallableGenericAlias = ta._CallableGenericAlias  # type: ignore  # noqa
_GenericAlias = ta._GenericAlias  # type: ignore  # noqa
_LiteralGenericAlias = ta._LiteralGenericAlias  # type: ignore  # noqa
_SpecialGenericAlias = ta._SpecialGenericAlias  # type: ignore  # noqa
_UnionGenericAlias = ta._UnionGenericAlias   # type: ignore  # noqa


##


@dc.dataclass(frozen=True)
class _Special:
    name: str
    alias: _SpecialGenericAlias  # type: ignore
    origin: type
    nparams: int


_KNOWN_SPECIALS = [
    _Special(
        v._name,  # noqa
        v,
        v.__origin__,
        v._nparams,  # noqa
    )
    for v in ta.__dict__.values()  # noqa
    if isinstance(v, _SpecialGenericAlias)
]

_KNOWN_SPECIALS_BY_NAME = {s.name: s for s in _KNOWN_SPECIALS}
_KNOWN_SPECIALS_BY_ALIAS = {s.alias: s for s in _KNOWN_SPECIALS}
_KNOWN_SPECIALS_BY_ORIGIN = {s.origin: s for s in _KNOWN_SPECIALS}

_MAX_KNOWN_SPECIAL_TYPE_VARS = 16

_KNOWN_SPECIAL_TYPE_VARS = tuple(
    ta.TypeVar(f'_{i}')  # noqa
    for i in range(_MAX_KNOWN_SPECIAL_TYPE_VARS)
)


##


_SIMPLE_GENERIC_ALIAS_TYPES: set[type] = set()


def is_simple_generic_alias_type(oty: type) -> bool:
    return (
        oty is _GenericAlias or
        oty is ta.GenericAlias or  # type: ignore  # noqa
        oty in _SIMPLE_GENERIC_ALIAS_TYPES
    )


def get_params(obj: ta.Any) -> tuple[ta.TypeVar, ...]:
    if isinstance(obj, type):
        if issubclass(obj, ta.Generic):  # type: ignore
            return obj.__dict__.get('__parameters__', ())  # noqa

        if (ks := _KNOWN_SPECIALS_BY_ORIGIN.get(obj)) is not None:
            if (np := ks.nparams) < 0:
                raise TypeError(obj)
            return _KNOWN_SPECIAL_TYPE_VARS[:np]

    oty = type(obj)

    if is_simple_generic_alias_type(oty):
        return obj.__dict__.get('__parameters__', ())  # noqa

    if oty is _CallableGenericAlias:
        raise NotImplementedError('get_params not yet implemented for typing.Callable')

    raise TypeError(obj)


def is_union_type(cls: ta.Any) -> bool:
    if hasattr(ta, 'UnionType'):
        return ta.get_origin(cls) in {ta.Union, getattr(ta, 'UnionType')}

    else:
        return ta.get_origin(cls) in {ta.Union}


def get_orig_bases(obj: ta.Any) -> ta.Any:
    return obj.__orig_bases__  # noqa


def get_orig_class(obj: ta.Any) -> ta.Any:
    return obj.__orig_class__  # noqa


def get_newtype_supertype(obj: ta.Any) -> ta.Any:
    return obj.__supertype__


def get_type_var_bound(obj: ta.Any) -> ta.Any:
    return obj.__bound__


##


class TypeInfo(abc.ABC):  # noqa
    pass


Type: ta.TypeAlias = ta.Union[  # noqa
    type,
    ta.TypeVar,
    TypeInfo,
]


TYPES: tuple[type, ...] = (
    type,
    ta.TypeVar,
    TypeInfo,
)


##


@dc.dataclass(frozen=True)
class Union(TypeInfo):
    args: frozenset[Type]

    @property
    def is_optional(self) -> bool:
        return _NoneType in self.args

    def without_none(self) -> Type:
        if _NoneType not in self.args:
            return self
        rem = self.args - _NONE_TYPE_FROZENSET
        if len(rem) == 1:
            return next(iter(rem))
        return Union(rem)


#


@dc.dataclass(frozen=True)
class Generic(TypeInfo):
    cls: type
    args: tuple[Type, ...]                                                # map[int, V] = (int, V) | map[T, T] = (T, T)

    params: tuple[ta.TypeVar, ...] = dc.field(compare=False, repr=False)  # map[int, V] = (_0, _1) | map[T, T] = (_0, _1)  # noqa
    # params2: tuple[ta.TypeVar, ...]                                     # map[int, V] = (V,)     | map[T, T] = (T,)

    obj: ta.Any = dc.field(compare=False, repr=False)

    # def __post_init__(self) -> None:
    #     if not isinstance(self.cls, type):
    #         raise TypeError(self.cls)

    def full_eq(self, other: 'Generic') -> bool:
        return (
            self.cls == other.cls and
            self.args == other.args and
            self.params == other.params and
            self.obj == other.obj
        )


#


@dc.dataclass(frozen=True)
class NewType(TypeInfo):
    obj: ta.Any
    ty: Type


#


@dc.dataclass(frozen=True)
class Annotated(TypeInfo):
    ty: Type
    md: ta.Sequence[ta.Any]

    obj: ta.Any = dc.field(compare=False, repr=False)


#


@dc.dataclass(frozen=True)
class Literal(TypeInfo):
    args: tuple[ta.Any, ...]

    obj: ta.Any = dc.field(compare=False, repr=False)


#


class Any(TypeInfo):
    pass


ANY = Any()


##


class ReflectTypeError(TypeError):
    pass


class Reflector:
    def __init__(
            self,
            *,
            override: ta.Callable[[ta.Any], ta.Any] | None = None,
    ) -> None:
        super().__init__()

        self._override = override

    #

    def is_type(self, obj: ta.Any) -> bool:
        try:
            self._type(obj, check_only=True)
        except ReflectTypeError:
            return False
        else:
            return True

    def type(self, obj: ta.Any) -> Type:
        if (ty := self._type(obj, check_only=False)) is None:
            raise RuntimeError(obj)
        return ty

    #

    def _type(self, obj: ta.Any, *, check_only: bool) -> Type | None:
        ##
        # Overrides beat everything

        if self._override is not None:
            if (ovr := self._override(obj)) is not None:
                return ovr

        ##
        # Any

        if obj is ta.Any:
            return ANY

        ##
        # Already a Type?

        if isinstance(obj, (ta.TypeVar, TypeInfo)):  # noqa
            return obj

        oty = type(obj)

        ##
        # Union

        if oty is _UnionGenericAlias or oty is types.UnionType:
            if check_only:
                return None

            return Union(frozenset(self.type(a) for a in ta.get_args(obj)))

        ##
        # NewType

        if isinstance(obj, ta.NewType):  # noqa
            if check_only:
                return None

            return NewType(obj, get_newtype_supertype(obj))

        ##
        # Simple Generic

        if (
                is_simple_generic_alias_type(oty) or
                oty is _CallableGenericAlias
        ):
            if check_only:
                return None

            origin = ta.get_origin(obj)
            args = ta.get_args(obj)

            if oty is _CallableGenericAlias:
                p, r = args
                if p is Ellipsis or isinstance(p, ta.ParamSpec):
                    raise ReflectTypeError(f'Callable argument not yet supported for {obj=}')
                args = (*p, r)
                params = _KNOWN_SPECIAL_TYPE_VARS[:len(args)]

            elif origin is tuple:
                params = _KNOWN_SPECIAL_TYPE_VARS[:len(args)]

            elif origin is ta.Generic:
                params = args

            else:
                params = get_params(origin)

            if len(args) != len(params):
                raise ReflectTypeError(f'Mismatched {args=} and {params=} for {obj=}')

            return Generic(
                origin,
                tuple(self.type(a) for a in args),
                params,
                obj,
            )

        ##
        # Full Generic

        if isinstance(obj, type):
            if check_only:
                return None

            if issubclass(obj, ta.Generic):  # type: ignore
                params = get_params(obj)
                if params:
                    return Generic(
                        obj,
                        params,
                        params,
                        obj,
                    )

            return obj

        ##
        # Special Generic

        if isinstance(obj, _SpecialGenericAlias):
            if (ks := _KNOWN_SPECIALS_BY_ALIAS.get(obj)) is not None:
                if check_only:
                    return None

                if (np := ks.nparams) < 0:
                    raise ReflectTypeError(obj)

                params = _KNOWN_SPECIAL_TYPE_VARS[:np]
                return Generic(
                    ks.origin,
                    params,
                    params,
                    obj,
                )

        ##
        # Annotated

        if isinstance(obj, _AnnotatedAlias):
            if check_only:
                return None

            o = ta.get_args(obj)[0]
            return Annotated(self.type(o), md=obj.__metadata__, obj=obj)

        ##
        # Literal

        if isinstance(obj, _LiteralGenericAlias):
            if check_only:
                return None

            return Literal(ta.get_args(obj), obj=obj)

        ##
        # Failure

        raise ReflectTypeError(obj)


DEFAULT_REFLECTOR = Reflector()

is_type = DEFAULT_REFLECTOR.is_type
type_ = DEFAULT_REFLECTOR.type



__version__ = '0.0.0.dev312'
__revision__ = '4714bf354df85d31b81de62365efd2756b378965'


#


class ProjectBase:
    name: str | None = None
    authors = [{'name': 'wrmsr'}]
    urls = {'source': 'https://github.com/wrmsr/omlish'}
    license = {'text': 'BSD-3-Clause'}
    requires_python = '>=3.12'

    version = __version__

    classifiers = [
        'License :: OSI Approved :: BSD License',
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',

        'Operating System :: OS Independent',
        'Operating System :: POSIX',
    ]


class Project(ProjectBase):
    name = 'omlish'
    description = 'omlish'

    #

    optional_dependencies = {
        'async': [
            'anyio ~= 4.9',
            'sniffio ~= 1.3',

            'greenlet ~= 3.2',

            'trio ~= 0.30',
            'trio-asyncio ~= 0.15',
        ],

        'compress': [
            'lz4 ~= 4.4',
            # 'lz4 @ git+https://github.com/wrmsr/python-lz4@wrmsr_20240830_GIL_NOT_USED'

            'python-snappy ~= 0.7',

            'zstandard ~= 0.23',

            'brotli ~= 1.1',
        ],

        'diag': [
            'asttokens ~= 3.0',
            'executing ~= 2.2',

            'psutil ~= 7.0',
        ],

        'formats': [
            'orjson ~= 3.10',
            'ujson ~= 5.10',

            'pyyaml ~= 6.0',

            'cbor2 ~= 5.6',

            'cloudpickle ~= 3.1',
        ],

        'http': [
            'httpx[http2] ~= 0.28',
        ],

        'misc': [
            'wrapt ~= 1.17',
        ],

        'secrets': [
            'cryptography ~= 44.0',
        ],

        'sqlalchemy': [
            'sqlalchemy[asyncio] ~= 2.0',
        ],

        'sqldrivers': [
            'pg8000 ~= 1.31',
            # 'psycopg2 ~= 2.9',
            # 'psycopg ~= 3.2',

            'pymysql ~= 1.1',
            # 'mysql-connector-python ~= 9.3',
            # 'mysqlclient ~= 2.2',

            'aiomysql ~= 0.2',
            'aiosqlite ~= 0.21',
            'asyncpg ~= 0.30',

            'apsw ~= 3.49',

            'sqlean.py ~= 3.49',

            'duckdb ~= 1.2',
        ],

        'templates': [
            'markupsafe ~= 3.0',

            'jinja2 ~= 3.1',
        ],

        'testing': [
            'pytest ~= 8.3',
        ],
    }

    #

    _plus_dependencies = [
        'anyio',
        'sniffio',

        'asttokens',
        'executing',

        'orjson',
        'pyyaml',

        'wrapt',
    ]

    _dependency_specs_by_name = (lambda od: {  # noqa
        s.split()[0]: s
        for l in od.values() for s in l
    })(optional_dependencies)

    optional_dependencies['plus'] = (lambda ds, pd: [  # noqa
        ds[n] for n in pd
    ])(_dependency_specs_by_name, _plus_dependencies)

    #

    entry_points = {
        'omlish.manifests': {name: name},
    }


#


class SetuptoolsBase:
    manifest_in = [
        'global-exclude **/conftest.py',
    ]

    find_packages = {
        'exclude': [
            '*.tests',
            '*.tests.*',
        ],
    }

    package_data = {
        '*': [
            '*.c',
            '*.cc',
            '*.cu',
            '*.g4',
            '*.h',

            '.manifests.json',

            'LICENSE',
            'LICENSE.txt',
        ],
    }


class Setuptools(SetuptoolsBase):
    find_packages = {
        'include': [Project.name, f'{Project.name}.*'],
        'exclude': [*SetuptoolsBase.find_packages['exclude']],
    }



import abc
import re
import typing as ta


##


class StringCasingError(Exception):
    pass


class ImproperStringCasingError(StringCasingError):
    pass


class UnknownStringCasingError(StringCasingError):
    pass


class AmbiguousStringCasingError(StringCasingError):
    pass


def _check_all_lowercase(*ps: str) -> None:
    for p in ps:
        if p.lower() != p:
            raise ImproperStringCasingError(p)


##


class StringCasing(abc.ABC):
    @abc.abstractmethod
    def match(self, s: str) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def split(self, s: str) -> list[str]:
        """Returns lowercase."""

        raise NotImplementedError

    @abc.abstractmethod
    def join(self, *parts: str) -> str:
        """Expects lowercase."""

        raise NotImplementedError


#


class CamelCase(StringCasing):
    """FooBarBaz"""

    _PAT: ta.ClassVar[re.Pattern] = re.compile(r'[A-Z][a-z0-9]*(?:[A-Z][a-z0-9]*)*')
    _SPLIT_PAT: ta.ClassVar[re.Pattern] = re.compile(r'[A-Z][a-z0-9]*')

    def match(self, s: str) -> bool:
        return bool(self._PAT.fullmatch(s))

    def split(self, s: str) -> list[str]:
        if not self.match(s):
            raise ImproperStringCasingError(f'Not valid CamelCase: {s!r}')
        return [m.group(0).lower() for m in self._SPLIT_PAT.finditer(s)]

    def join(self, *parts: str) -> str:
        _check_all_lowercase(*parts)
        return ''.join(p.capitalize() for p in parts)


#


class LowCamelCase(StringCasing):
    """fooBarBaz"""

    _MATCH_PAT: ta.ClassVar[re.Pattern] = re.compile(r'[a-z][a-z0-9]*(?:[A-Z][a-z0-9]*)*')
    _FIRST_PAT: ta.ClassVar[re.Pattern] = re.compile(r'^[a-z0-9]+')
    _UPPER_PAT: ta.ClassVar[re.Pattern] = re.compile(r'[A-Z][a-z0-9]*')

    def match(self, s: str) -> bool:
        return bool(self._MATCH_PAT.fullmatch(s))

    def split(self, s: str) -> list[str]:
        if not self.match(s):
            raise ImproperStringCasingError(f'Not valid lowCamelCase: {s!r}')
        parts: list[str] = []
        m0 = self._FIRST_PAT.match(s)
        if m0:
            parts.append(m0.group(0))
            start = m0.end()
        else:
            start = 0
        for m in self._UPPER_PAT.finditer(s, pos=start):
            parts.append(m.group(0))
        return [p.lower() for p in parts]

    def join(self, *parts: str) -> str:
        _check_all_lowercase(*parts)
        if not parts:
            return ''
        first, *rest = parts
        return first.lower() + ''.join(p.capitalize() for p in rest)


#


class SnakeCase(StringCasing):
    """foo_bar_baz"""

    _PAT: ta.ClassVar[re.Pattern] = re.compile(r'[a-z0-9]+(?:_[a-z0-9]+)*')

    def match(self, s: str) -> bool:
        return bool(self._PAT.fullmatch(s))

    def split(self, s: str) -> list[str]:
        if not self.match(s):
            raise ImproperStringCasingError(f'Not valid snake_case: {s!r}')
        return s.split('_')

    def join(self, *parts: str) -> str:
        _check_all_lowercase(*parts)
        return '_'.join(p.lower() for p in parts)


#


class UpSnakeCase(StringCasing):
    """FOO_BAR_BAZ"""

    _PAT: ta.ClassVar[re.Pattern] = re.compile(r'[A-Z0-9]+(?:_[A-Z0-9]+)*')

    def match(self, s: str) -> bool:
        return bool(self._PAT.fullmatch(s))

    def split(self, s: str) -> list[str]:
        if not self.match(s):
            raise ImproperStringCasingError(f'Not valid UPPER_SNAKE_CASE: {s!r}')
        return [part.lower() for part in s.split('_')]

    def join(self, *parts: str) -> str:
        _check_all_lowercase(*parts)
        return '_'.join(p.upper() for p in parts)


##


STRING_CASINGS = [
    (CAMEL_CASE := CamelCase()),
    (LOW_CAMEL_CASE := LowCamelCase()),
    (SNAKE_CASE := SnakeCase()),
    (UP_SNAKE_CASE := UpSnakeCase()),
]


camel_case = CAMEL_CASE.join
low_camel_case = LOW_CAMEL_CASE.join
snake_case = SNAKE_CASE.join
up_snake_case = UP_SNAKE_CASE.join


##


def get_string_casing(s: str, casings: ta.Iterable[StringCasing] | None = None) -> StringCasing:
    if casings is None:
        casings = STRING_CASINGS
    cs = [c for c in casings if c.match(s)]
    if not cs:
        raise UnknownStringCasingError
    if len(cs) != 1:
        raise AmbiguousStringCasingError
    [c] = cs
    return c


def split_string_casing(s: str, casings: ta.Iterable[StringCasing] | None = None) -> list[str]:
    if casings is None:
        casings = STRING_CASINGS
    ts = {tuple(c.split(s)) for c in casings if c.match(s)}
    if not ts:
        raise UnknownStringCasingError
    if len(ts) != 1:
        raise AmbiguousStringCasingError
    [t] = ts
    return list(t)



import dataclasses as dc
import typing as ta

from .... import lang
from ....funcs.genmachine import GenMachine
from .errors import JsonStreamError
from .lexing import SCALAR_VALUE_TYPES
from .lexing import VALUE_TOKEN_KINDS
from .lexing import Position
from .lexing import ScalarValue
from .lexing import Token


##


class BeginObject(lang.Marker):
    pass


class Key(ta.NamedTuple):
    key: str


class EndObject(lang.Marker):
    pass


class BeginArray(lang.Marker):
    pass


class EndArray(lang.Marker):
    pass


JsonStreamParserEvent: ta.TypeAlias = ta.Union[  # noqa
    type[BeginObject],
    Key,
    type[EndObject],

    type[BeginArray],
    type[EndArray],

    ScalarValue,
]


class JsonStreamParserEvents(lang.Namespace):
    BeginObject = BeginObject
    Key = Key
    EndObject = EndObject

    BeginArray = BeginArray
    EndArray = EndArray


##


def yield_parser_events(obj: ta.Any) -> ta.Generator[JsonStreamParserEvent, None, None]:
    if isinstance(obj, SCALAR_VALUE_TYPES):
        yield obj  # type: ignore

    elif isinstance(obj, ta.Mapping):
        yield BeginObject
        for k, v in obj.items():
            yield Key(k)
            yield from yield_parser_events(v)
        yield EndObject

    elif isinstance(obj, ta.Sequence):
        yield BeginArray
        for v in obj:
            yield from yield_parser_events(v)
        yield EndArray

    else:
        raise TypeError(obj)


##


@dc.dataclass()
class JsonStreamParseError(JsonStreamError):
    message: str

    pos: Position | None = None


class JsonStreamObject(list):
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({super().__repr__()})'


class JsonStreamParser(GenMachine[Token, JsonStreamParserEvent]):
    def __init__(self) -> None:
        super().__init__(self._do_value())

        self._stack: list[ta.Literal['OBJECT', 'KEY', 'ARRAY']] = []

    #

    def _next_tok(self):
        while True:
            tok = yield None

            if tok.kind != 'SPACE':
                return tok

    #

    def _emit_event(self, v):
        if not self._stack:
            return ((v,), self._do_value())

        tt = self._stack[-1]
        if tt == 'KEY':
            self._stack.pop()
            if not self._stack:
                raise JsonStreamParseError('Unexpected key')

            tt2 = self._stack[-1]
            if tt2 == 'OBJECT':
                return ((v,), self._do_after_pair())

            else:
                raise JsonStreamParseError('Unexpected key')

        elif tt == 'ARRAY':
            return ((v,), self._do_after_element())

        else:
            raise JsonStreamParseError(f'Unexpected value: {v!r}')

    #

    def _do_value(self, *, must_be_present: bool = False):
        try:
            tok = yield from self._next_tok()
        except GeneratorExit:
            if self._stack:
                raise JsonStreamParseError('Expected value') from None
            else:
                raise

        if tok.kind in VALUE_TOKEN_KINDS:
            y, r = self._emit_event(tok.value)
            yield y
            return r

        elif tok.kind == 'LBRACE':
            y, r = self._emit_begin_object()
            yield y
            return r

        elif tok.kind == 'LBRACKET':
            y, r = self._emit_begin_array()
            yield y
            return r

        elif must_be_present:
            raise JsonStreamParseError('Expected value', tok.pos)

        elif tok.kind == 'RBRACKET':
            y, r = self._emit_end_array()
            yield y
            return r

        else:
            raise JsonStreamParseError('Expected value', tok.pos)

    #

    def _emit_begin_object(self):
        self._stack.append('OBJECT')
        return ((BeginObject,), self._do_object_body())

    def _emit_end_object(self):
        if not self._stack:
            raise JsonStreamParseError('Unexpected end object')

        tt = self._stack.pop()
        if tt != 'OBJECT':
            raise JsonStreamParseError('Unexpected end object')

        return self._emit_event(EndObject)

    def _do_object_body(self, *, must_be_present: bool = False):
        try:
            tok = yield from self._next_tok()
        except GeneratorExit:
            raise JsonStreamParseError('Expected object body') from None

        if tok.kind == 'STRING':
            k = tok.value

            try:
                tok = yield from self._next_tok()
            except GeneratorExit:
                raise JsonStreamParseError('Expected key') from None
            if tok.kind != 'COLON':
                raise JsonStreamParseError('Expected colon', tok.pos)

            yield (Key(k),)
            self._stack.append('KEY')
            return self._do_value()

        elif must_be_present:
            raise JsonStreamParseError('Expected value', tok.pos)

        elif tok.kind == 'RBRACE':
            y, r = self._emit_end_object()
            yield y
            return r

        else:
            raise JsonStreamParseError('Expected value', tok.pos)

    def _do_after_pair(self):
        try:
            tok = yield from self._next_tok()
        except GeneratorExit:
            raise JsonStreamParseError('Expected continuation') from None

        if tok.kind == 'COMMA':
            return self._do_object_body(must_be_present=True)

        elif tok.kind == 'RBRACE':
            y, r = self._emit_end_object()
            yield y
            return r

        else:
            raise JsonStreamParseError('Expected continuation', tok.pos)

    #

    def _emit_begin_array(self):
        self._stack.append('ARRAY')
        return ((BeginArray,), self._do_value())

    def _emit_end_array(self):
        if not self._stack:
            raise JsonStreamParseError('Expected end array')

        tt = self._stack.pop()
        if tt != 'ARRAY':
            raise JsonStreamParseError('Unexpected end array')

        return self._emit_event(EndArray)

    def _do_after_element(self):
        try:
            tok = yield from self._next_tok()
        except GeneratorExit:
            raise JsonStreamParseError('Expected continuation') from None

        if tok.kind == 'COMMA':
            return self._do_value(must_be_present=True)

        elif tok.kind == 'RBRACKET':
            y, r = self._emit_end_array()
            yield y
            return r

        else:
            raise JsonStreamParseError('Expected continuation', tok.pos)



# ruff: noqa: I001
# fmt: off
import pytest

from . import (  # noqa
    asyncs,
    depskip,
    logging,
    managermarks,
    pydevd,
    repeat,
    skips,
    spacing,
    switches,
)

from ._registry import (  # noqa
    ALL,
    register,
)


def add_hooks(pm: pytest.PytestPluginManager) -> None:
    present_types = {type(p) for p in pm.get_plugins()}

    for plugin in ALL:
        if plugin not in present_types:
            pm.register(plugin())  # noqa



# @omlish-lite
import shlex


##


_DEFAULT_DOCKER_TIMEBOMB_NAME = 'omlish-timebomb'


def docker_timebomb_payload(delay_s: float, name: str = _DEFAULT_DOCKER_TIMEBOMB_NAME) -> str:
    return (
        '('
        f'echo {shlex.quote(name)} && '
        f'sleep {delay_s:g} && '
        'sh -c \'killall5 -9 -o $PPID -o $$ ; kill 1\''
        ') &'
    )



# ruff: noqa: UP006 UP007
# @omlish-lite
import dataclasses as dc
import typing as ta


NameAliasesManifestT = ta.TypeVar('NameAliasesManifestT', bound='NameAliasesManifest')


##


@dc.dataclass(frozen=True)
class ModAttrManifest:
    mod_name: str
    attr_name: str

    def load(self) -> ta.Any:
        importlib = __import__('importlib')
        mod = importlib.import_module(self.mod_name)
        return getattr(mod, self.attr_name)


##


@dc.dataclass(frozen=True)
class NameAliasesManifest:
    name: str
    aliases: ta.Optional[ta.Collection[str]] = None

    @classmethod
    def build_name_dict(cls, objs: ta.Iterable[NameAliasesManifestT]) -> ta.Dict[str, NameAliasesManifestT]:
        dct: ta.Dict[str, NameAliasesManifestT] = {}
        for o in objs:
            for n in (o.name, *(o.aliases or ())):
                if n in dct:
                    raise KeyError(n)
                dct[n] = o
        return dct
